# SPDX-FileCopyrightText: © 2025 DSLab - Fondazione Bruno Kessler
#
# SPDX-License-Identifier: Apache-2.0

from __future__ import annotations

from digitalhub.entities._base.material.status import MaterialStatus


class DataitemStatus(MaterialStatus):
    """
    DataitemStatus status.
    """

    def __init__(
        self,
        state: str,
        message: str | None = None,
        files: list[dict] | None = None,
        preview: dict | None = None,
        **kwargs,
    ) -> None:
        super().__init__(state, message, files)
        self.preview = preview



# SPDX-FileCopyrightText: © 2025 DSLab - Fondazione Bruno Kessler
#
# SPDX-License-Identifier: Apache-2.0

from __future__ import annotations

import typing

from digitalhub.utils.exceptions import BuilderError

if typing.TYPE_CHECKING:
    from digitalhub.runtimes._base import Runtime


class RuntimeBuilder:
    """
    Builder class for instantiating runtime objects.

    This class implements the Builder pattern to create Runtime instances.
    Subclasses must set the RUNTIME_CLASS class variable to specify which
    Runtime implementation to build.

    Attributes
    ----------
    RUNTIME_CLASS : Runtime
        The Runtime class to be instantiated by this builder.

    Raises
    ------
    BuilderError
        If RUNTIME_CLASS is not set in the implementing class.
    """

    RUNTIME_CLASS: Runtime = None

    def __init__(self) -> None:
        """
        Initialize a RuntimeBuilder instance.

        Raises
        ------
        BuilderError
            If RUNTIME_CLASS is not set in the implementing class.
        """
        if self.RUNTIME_CLASS is None:
            raise BuilderError("RUNTIME_CLASS must be set")

    def build(self, project: str, *args, **kwargs) -> Runtime:
        """
        Build a runtime object.

        Parameters
        ----------
        project : str
            The project identifier for the runtime instance.
        *args
            Additional positional arguments to pass to the Runtime constructor.
        **kwargs
            Additional keyword arguments to pass to the Runtime constructor.

        Returns
        -------
        Runtime
            A new instance of the configured Runtime class.
        """
        return self.RUNTIME_CLASS(project, *args, **kwargs)



# SPDX-FileCopyrightText: © 2025 DSLab - Fondazione Bruno Kessler
#
# SPDX-License-Identifier: Apache-2.0

from __future__ import annotations

from digitalhub.entities.model._base.spec import ModelSpec, ModelValidator


class ModelSpecSklearn(ModelSpec):
    """
    ModelSpecSklearn specifications.
    """


class ModelValidatorSklearn(ModelValidator):
    """
    ModelValidatorSklearn validator.
    """



# SPDX-FileCopyrightText: © 2025 DSLab - Fondazione Bruno Kessler
#
# SPDX-License-Identifier: Apache-2.0

from __future__ import annotations

import typing
from typing import Any

from digitalhub.entities._base.material.utils import refresh_decorator
from digitalhub.entities.dataitem._base.entity import Dataitem
from digitalhub.stores.data.api import get_store
from digitalhub.utils.uri_utils import has_sql_scheme

if typing.TYPE_CHECKING:
    from digitalhub.entities._base.entity.metadata import Metadata
    from digitalhub.entities.dataitem.table.spec import DataitemSpecTable
    from digitalhub.entities.dataitem.table.status import DataitemStatusTable


class DataitemTable(Dataitem):
    """
    DataitemTable class.
    """

    def __init__(
        self,
        project: str,
        name: str,
        uuid: str,
        kind: str,
        metadata: Metadata,
        spec: DataitemSpecTable,
        status: DataitemStatusTable,
        user: str | None = None,
    ) -> None:
        super().__init__(project, name, uuid, kind, metadata, spec, status, user)

        self.spec: DataitemSpecTable
        self.status: DataitemStatusTable

        self._query: str | None = None

    def query(self, query: str) -> DataitemTable:
        """
        Set query to execute.

        Parameters
        ----------
        query : str
            Query to execute.

        Returns
        -------
        DataitemTable
            Self object.
        """
        # to remove in future
        if not has_sql_scheme(self.spec.path):
            raise ValueError(
                f"Dataitem path is not a SQL scheme: {self.spec.path}",
                " Query can be made only on a SQL scheme.",
            )
        self._query = query
        return self

    @refresh_decorator
    def as_df(
        self,
        file_format: str | None = None,
        engine: str | None = "pandas",
        **kwargs,
    ) -> Any:
        """
        Read dataitem file (csv or parquet) as a DataFrame from spec.path.
        It's possible to pass additional arguments to the this function. These
        keyword arguments will be passed to the DataFrame reader function such as
        pandas's read_csv or read_parquet.

        Parameters
        ----------
        file_format : str
            Format of the file to read. By default, it will be inferred from
            the extension of the file.
        engine : str
            Dataframe framework, by default pandas.
        **kwargs : dict
            Keyword arguments passed to the read_df function.

        Returns
        -------
        Any
            DataFrame.
        """
        if self._query is not None:
            df = get_store(self.spec.path).query(
                self._query,
                self.spec.path,
                engine,
            )
            self._query = None
            return df
        return get_store(self.spec.path).read_df(
            self.spec.path,
            file_format,
            engine,
            **kwargs,
        )

    @refresh_decorator
    def write_df(
        self,
        df: Any,
        extension: str | None = None,
        **kwargs,
    ) -> str:
        """
        Write DataFrame as parquet/csv/table into dataitem spec.path.
        keyword arguments will be passed to the DataFrame reader function such as
        pandas's to_csv or to_parquet.
        Note that by default the index is dropped when writing the dataframe. To
        keep the index, you can pass index=True as a keyword argument.
        If the dataitem path is a SQL scheme, the dataframe will be written to the
        table specified in the path (sql://<database_name>(/<schema_name>)/<table_name>).

        Parameters
        ----------
        df : Any
            DataFrame to write.
        extension : str
            Extension of the file (supported parquet and csv).
        **kwargs : dict
            Keyword arguments passed to the write_df function.

        Returns
        -------
        str
            Path to the written dataframe.

        Examples
        --------
        >>> import digitalhub as dh
        >>> import pandas as pd
        >>>
        >>> p = dh.get_project("my_project")
        >>> df = pd.read_df("data/my_data.csv")
        >>> di = p.new_dataitem(
        ...     name="my_dataitem",
        ...     kind="table",
        ...     path="s3://my-bucket/my-data.parquet",
        ... )
        >>> di.write_df(
        ...     df,
        ...     extension="parquet",
        ...     index=True,
        ... )
        's3://my-bucket/my-data.parquet'
        """
        return get_store(self.spec.path).write_df(
            df,
            self.spec.path,
            extension=extension,
            **kwargs,
        )



# SPDX-FileCopyrightText: © 2025 DSLab - Fondazione Bruno Kessler
#
# SPDX-License-Identifier: Apache-2.0

from __future__ import annotations

import typing

from digitalhub.entities._commons.enums import EntityTypes
from digitalhub.entities._processors.processors import context_processor

if typing.TYPE_CHECKING:
    from digitalhub.entities.trigger._base.entity import Trigger


ENTITY_TYPE = EntityTypes.TRIGGER.value


def new_trigger(
    project: str,
    name: str,
    kind: str,
    task: str,
    function: str | None = None,
    workflow: str | None = None,
    uuid: str | None = None,
    description: str | None = None,
    labels: list[str] | None = None,
    embedded: bool = False,
    template: dict | None = None,
    **kwargs,
) -> Trigger:
    """
    Create a new object.

    Parameters
    ----------
    project : str
        Project name.
    name : str
        Object name.
    kind : str
        Kind the object.
    task : str
        Task string.
    function : str
        Function string.
    workflow : str
        Workflow string.
    uuid : str
        ID of the object.
    description : str
        Description of the object (human readable).
    labels : list[str]
        List of labels.
    embedded : bool
        Flag to determine if object spec must be embedded in project spec.
    **kwargs : dict
        Spec keyword arguments.

    Returns
    -------
    Trigger
        Object instance.

    Examples
    --------
    >>> obj = new_trigger(project="my-project",
    >>>                   kind="trigger",
    >>>                   name="my-trigger",)
    """
    if workflow is None:
        if function is None:
            raise ValueError("Workflow or function must be provided")
        executable_type = "function"
        executable = function
    else:
        executable_type = "workflow"
        executable = workflow

    # Prepare kwargs
    if kwargs is None:
        kwargs = {}
    kwargs["task"] = task
    kwargs[executable_type] = executable

    # Template handling
    if template is None:
        template = {}
    if not isinstance(template, dict):
        raise ValueError("Template must be a dictionary")
    template["task"] = task
    template[executable_type] = executable
    template["local_execution"] = False
    kwargs["template"] = template

    return context_processor.create_context_entity(
        project=project,
        name=name,
        kind=kind,
        uuid=uuid,
        description=description,
        labels=labels,
        embedded=embedded,
        entity_type=ENTITY_TYPE,
        **kwargs,
    )


def get_trigger(
    identifier: str,
    project: str | None = None,
    entity_id: str | None = None,
) -> Trigger:
    """
    Get object from backend.

    Parameters
    ----------
    identifier : str
        Entity key (store://...) or entity name.
    project : str
        Project name.
    entity_id : str
        Entity ID.

    Returns
    -------
    Trigger
        Object instance.

    Examples
    --------
    Using entity key:
    >>> obj = get_trigger("store://my-trigger-key")

    Using entity name:
    >>> obj = get_trigger("my-trigger-name"
    >>>                  project="my-project",
    >>>                  entity_id="my-trigger-id")
    """
    return context_processor.read_context_entity(
        identifier=identifier,
        entity_type=ENTITY_TYPE,
        project=project,
        entity_id=entity_id,
    )


def list_triggers(
    project: str,
    q: str | None = None,
    name: str | None = None,
    kind: str | None = None,
    user: str | None = None,
    state: str | None = None,
    created: str | None = None,
    updated: str | None = None,
    versions: str | None = None,
    task: str | None = None,
) -> list[Trigger]:
    """
    List all latest version objects from backend.

    Parameters
    ----------
    project : str
        Project name.
    q : str
        Query string to filter objects.
    name : str
        Object name.
    kind : str
        Kind of the object.
    user : str
        User that created the object.
    state : str
        Object state.
    created : str
        Creation date filter.
    updated : str
        Update date filter.
    versions : str
        Object version, default is latest.
    task : str
        Task string filter.

    Returns
    -------
    list[Trigger]
        List of object instances.

    Examples
    --------
    >>> objs = list_triggers(project="my-project")
    """
    return context_processor.list_context_entities(
        project=project,
        entity_type=ENTITY_TYPE,
        q=q,
        name=name,
        kind=kind,
        user=user,
        state=state,
        created=created,
        updated=updated,
        versions=versions,
        task=task,
    )


def import_trigger(
    file: str | None = None,
    key: str | None = None,
    reset_id: bool = False,
    context: str | None = None,
) -> Trigger:
    """
    Import an object from a YAML file or from a storage key.

    Parameters
    ----------
    file : str
        Path to the YAML file.
    key : str
        Entity key (store://...).
    reset_id : bool
        Flag to determine if the ID of executable entities should be reset.
    context : str
        Project name to use for context resolution.

    Returns
    -------
    Trigger
        Object instance.

    Examples
    --------
    >>> obj = import_trigger("my-trigger.yaml")
    """
    return context_processor.import_context_entity(
        file,
        key,
        reset_id,
        context,
    )


def load_trigger(file: str) -> Trigger:
    """
    Load object from a YAML file and update an existing object into the backend.

    Parameters
    ----------
    file : str
        Path to YAML file.

    Returns
    -------
    Trigger
        Object instance.

    Examples
    --------
    >>> obj = load_trigger("my-trigger.yaml")
    """
    return context_processor.load_context_entity(file)


def update_trigger(entity: Trigger) -> Trigger:
    """
    Update object. Note that object spec are immutable.

    Parameters
    ----------
    entity : Trigger
        Object to update.

    Returns
    -------
    Trigger
        Entity updated.

    Examples
    --------
    >>> obj = update_trigger(obj)
    """
    return context_processor.update_context_entity(
        project=entity.project,
        entity_type=entity.ENTITY_TYPE,
        entity_id=entity.id,
        entity_dict=entity.to_dict(),
    )


def delete_trigger(
    identifier: str,
    project: str | None = None,
    entity_id: str | None = None,
    delete_all_versions: bool = False,
) -> dict:
    """
    Delete object from backend.

    Parameters
    ----------
    identifier : str
        Entity key (store://...) or entity name.
    project : str
        Project name.
    entity_id : str
        Entity ID.
    delete_all_versions : bool
        Delete all versions of the named entity.
        If True, use entity name instead of entity key as identifier.

    Returns
    -------
    dict
        Response from backend.

    Examples
    --------
    If delete_all_versions is False:
    >>> obj = delete_trigger("store://my-trigger-key")

    Otherwise:
    >>> obj = delete_trigger("my-trigger-name"
    >>>                     project="my-project",
    >>>                     delete_all_versions=True)
    """
    return context_processor.delete_context_entity(
        identifier=identifier,
        entity_type=ENTITY_TYPE,
        project=project,
        entity_id=entity_id,
        delete_all_versions=delete_all_versions,
    )



# SPDX-FileCopyrightText: © 2025 DSLab - Fondazione Bruno Kessler
#
# SPDX-License-Identifier: Apache-2.0

from __future__ import annotations

import typing

from digitalhub.entities._commons.enums import EntityTypes
from digitalhub.entities._commons.utils import build_log_path_from_filename, build_log_path_from_source
from digitalhub.entities._constructors.uuid import build_uuid
from digitalhub.stores.data.api import get_store
from digitalhub.stores.readers.data.api import get_reader_by_engine, get_reader_by_object
from digitalhub.utils.file_utils import eval_local_source

if typing.TYPE_CHECKING:
    from digitalhub.entities.dataitem._base.entity import Dataitem
    from digitalhub.utils.types import Dataframe, SourcesOrListOfSources


def read_data_sample(
    source: SourcesOrListOfSources,
    file_format: str | None = None,
    read_df_params: dict | None = None,
    engine: str | None = None,
) -> Dataframe:  # type: ignore
    """
    Evaluate and load data from source or return provided data.

    Parameters
    ----------
    source : SourcesOrListOfSources
        The source specification(s) to load data from.
    file_format : str
        The file format specification for reading the source (e.g., 'parquet', 'csv').
    engine : str
        The engine to use for reading the data (e.g., 'pandas', 'polars').

    Returns
    -------
    Dataframe
        Loaded data object.
    """
    eval_local_source(source)
    if read_df_params is None or not isinstance(read_df_params, dict):
        read_df_params = {}
    reader = get_reader_by_engine(engine)
    read_df_params[reader.get_limit_arg_name()] = 10
    return get_store(source).read_df(
        source,
        file_format=file_format,
        engine=engine,
        **read_df_params,
    )


def process_data_kwargs(
    project: str,
    name: str,
    data: Dataframe,  # type: ignore
    path: str | None = None,
    source: SourcesOrListOfSources | None = None,
    filename: str | None = None,
    **kwargs,
) -> dict:
    """
    Process and enhance specification parameters for dataitem creation.

    Parameters
    ----------
    project : str
        The name of the project.
    name : str
        The name of the dataitem entity.
    kind : str
        The kind of dataitem being created (e.g., 'table').
    path : str
        The destination path for the dataitem entity.
        If None, a path will be automatically generated.
    data : Dataframe
        The data object for schema extraction and processing.
        Used as an alternative to source for table dataitems.
    source
    **kwargs : dict
        Additional specification parameters to be processed
        and passed to the dataitem creation.

    Returns
    -------
    dict
        The enhanced specification parameters for dataitem creation.
    """
    if (source is None) == (filename is None):
        raise ValueError("Either source or filename must be provided.")
    if source is None:
        fn = build_log_path_from_filename
        arg = filename
    else:
        fn = build_log_path_from_source
        arg = source

    if path is None:
        uuid = build_uuid()
        kwargs["uuid"] = uuid
        kwargs["path"] = fn(project, EntityTypes.DATAITEM.value, name, uuid, arg)

    kwargs["schema"] = get_reader_by_object(data).get_schema(data)
    return kwargs


def post_process(
    obj: Dataitem,
    data: Dataframe,  # type: ignore
) -> Dataitem:
    """
    Post-process dataitem object with additional metadata and previews.

    Enhances the dataitem object with additional information extracted
    from the data. For table dataitems, generates and stores a data
    preview in the object's status.

    Parameters
    ----------
    obj : Dataitem
        The dataitem object to post-process and enhance.
    data : Dataframe
        The data object used to generate previews and extract
        additional metadata information.

    Returns
    -------
    Dataitem
        The enhanced dataitem object with updated status information
        and saved changes.
    """
    reader = get_reader_by_object(data)
    obj.status.preview = reader.get_preview(data)
    obj.save(update=True)
    return obj



# SPDX-FileCopyrightText: © 2025 DSLab - Fondazione Bruno Kessler
#
# SPDX-License-Identifier: Apache-2.0

from __future__ import annotations

from digitalhub.entities._commons.enums import EntityKinds
from digitalhub.entities.model._base.builder import ModelBuilder
from digitalhub.entities.model.huggingface.entity import ModelHuggingface
from digitalhub.entities.model.huggingface.spec import ModelSpecHuggingface, ModelValidatorHuggingface
from digitalhub.entities.model.huggingface.status import ModelStatusHuggingface


class ModelHuggingfaceBuilder(ModelBuilder):
    """
    ModelHuggingface builder.
    """

    ENTITY_CLASS = ModelHuggingface
    ENTITY_SPEC_CLASS = ModelSpecHuggingface
    ENTITY_SPEC_VALIDATOR = ModelValidatorHuggingface
    ENTITY_STATUS_CLASS = ModelStatusHuggingface
    ENTITY_KIND = EntityKinds.MODEL_HUGGINGFACE.value



# SPDX-FileCopyrightText: © 2025 DSLab - Fondazione Bruno Kessler
#
# SPDX-License-Identifier: Apache-2.0

from __future__ import annotations

from pydantic import BaseModel


class Signature(BaseModel):
    """
    MLFlow model signature.
    """

    inputs: str | None = None
    outputs: str | None = None
    params: str | None = None

    def to_dict(self):
        return self.model_dump()


class Dataset(BaseModel):
    """
    MLFlow model dataset.
    """

    name: str | None = None
    digest: str | None = None
    profile: str | None = None
    dataset_schema: str | None = None
    source: str | None = None
    source_type: str | None = None

    def to_dict(self):
        return self.model_dump()



# SPDX-FileCopyrightText: © 2025 DSLab - Fondazione Bruno Kessler
#
# SPDX-License-Identifier: Apache-2.0

from __future__ import annotations

import typing

from digitalhub.entities._base.context.entity import ContextEntity
from digitalhub.entities._commons.metrics import MetricType, set_metrics, validate_metric_value
from digitalhub.entities._processors.processors import context_processor

if typing.TYPE_CHECKING:
    from digitalhub.entities._base.entity.metadata import Metadata
    from digitalhub.entities._base.entity.spec import Spec
    from digitalhub.entities._base.metrics.status import MetricsStatus


class MetricsEntity(ContextEntity):
    """
    Mixin class for entities with metrics.
    """

    def __init__(
        self,
        project: str,
        kind: str,
        metadata: Metadata,
        spec: Spec,
        status: MetricsStatus,
        user: str | None = None,
    ) -> None:
        super().__init__(project, kind, metadata, spec, status, user)
        self.status: MetricsStatus

    @property
    def metrics(self) -> dict[str, MetricType]:
        """
        Get metrics from entity status.

        Returns
        -------
        dict[str, MetricType]
            Metrics dictionary.
        """
        return self.status.metrics

    def log_metric(
        self,
        key: str,
        value: MetricType,
        overwrite: bool = False,
        single_value: bool = False,
    ) -> None:
        """
        Log metric into entity status.
        A metric is named by a key and value (single number or list of numbers).
        The metric by default is put in a list or appended to an existing list.
        If single_value is True, the value will be a single number.

        Parameters
        ----------
        key : str
            Key of the metric.
        value : MetricType
            Value of the metric.
        overwrite : bool
            If True, overwrite existing metric.
        single_value : bool
            If True, value is a single value.

        Examples
        --------
        Log a new value in a list
        >>> entity.log_metric("loss", 0.002)

        Append a new value in a list
        >>> entity.log_metric("loss", 0.0019)

        Log a list of values and append them to existing metric:
        >>> entity.log_metric(
        ...     "loss",
        ...     [
        ...         0.0018,
        ...         0.0015,
        ...     ],
        ... )

        Log a single value (not represented as list):
        >>> entity.log_metric(
        ...     "accuracy",
        ...     0.9,
        ...     single_value=True,
        ... )

        Log a list of values and overwrite existing metric:
        >>> entity.log_metric(
        ...     "accuracy",
        ...     [0.8, 0.9],
        ...     overwrite=True,
        ... )
        """
        self._set_metrics(key, value, overwrite, single_value)
        context_processor.update_metric(self.project, self.ENTITY_TYPE, self.id, key, self.status.metrics[key])

    def log_metrics(
        self,
        metrics: dict[str, MetricType],
        overwrite: bool = False,
    ) -> None:
        """
        Log metrics into entity status. If a metric is a list, it will be logged as a list.
        Otherwise, it will be logged as a single value.

        Parameters
        ----------
        metrics : dict[str, MetricType]
            Dict of metrics to log.
        overwrite : bool
            If True, overwrite existing metrics.

        Examples
        --------
        Log multiple metrics at once
        >>> entity.log_metrics(
        ...     {
        ...         "loss": 0.002,
        ...         "accuracy": 0.95,
        ...     }
        ... )

        Log metrics with lists and single values
        >>> entity.log_metrics(
        ...     {
        ...         "loss": [
        ...             0.1,
        ...             0.05,
        ...         ],
        ...         "epoch": 10,
        ...     }
        ... )

        Append to existing metrics (default behavior)
        >>> entity.log_metrics(
        ...     {
        ...         "loss": 0.001,
        ...         "accuracy": 0.96,
        ...     }
        ... )  # Appends to existing

        Overwrite existing metrics
        >>> entity.log_metrics(
        ...     {
        ...         "loss": 0.0005,
        ...         "accuracy": 0.98,
        ...     },
        ...     overwrite=True,
        ... )

        See also
        --------
        log_metric
        """
        for key, value in metrics.items():
            # For lists, use log_metric which handles appending correctly
            if isinstance(value, list):
                self.log_metric(key, value, overwrite)

            # For single values, check if we should append or create new
            else:
                if not overwrite and key in self.status.metrics:
                    self.log_metric(key, value)
                else:
                    self.log_metric(key, value, overwrite, single_value=True)

    def _post_read_hook(self) -> None:
        """
        Hook method called after reading the entity from Core.
        Can be overridden in subclasses to implement custom behavior.
        """
        self.read_metrics()

    def read_metrics(self) -> None:
        """
        Get model metrics from backend.
        """
        self.status.metrics = context_processor.read_metrics(
            project=self.project,
            entity_type=self.ENTITY_TYPE,
            entity_id=self.id,
        )

    def _set_metrics(
        self,
        key: str,
        value: MetricType,
        overwrite: bool,
        single_value: bool,
    ) -> None:
        """
        Set model metrics.

        Parameters
        ----------
        key : str
            Key of the metric.
        value : MetricType
            Value of the metric.
        overwrite : bool
            If True, overwrite existing metric.
        single_value : bool
            If True, value is a single value.
        """
        value = validate_metric_value(value)
        self.status.metrics = set_metrics(
            self.status.metrics,
            key,
            value,
            overwrite,
            single_value,
        )



# SPDX-FileCopyrightText: © 2025 DSLab - Fondazione Bruno Kessler
#
# SPDX-License-Identifier: Apache-2.0

from __future__ import annotations

from digitalhub.entities._commons.utils import build_log_path_from_source
from digitalhub.entities._constructors.uuid import build_uuid


def build_log_kwargs(
    project: str,
    name: str,
    entity_type: str,
    source: str | list[str],
    path: str | None = None,
    **kwargs,
) -> dict:
    """
    Build and enhance specification parameters for entity creation.

    Parameters
    ----------
    project : str
        The name of the project containing the entity.
    name : str
        The name of the entity.
    entity_type : str
        The type of the entity being created.
    source : str or list[str]
        The source specification(s) for the entity content.
        Can be a single source or multiple sources.
    path : str
        The destination path for the entity.
        If None, a path will be automatically generated.
    **kwargs : dict
        Additional specification parameters to be processed
        and passed to the entity creation.

    Returns
    -------
    dict
        The updated kwargs dictionary with processed path
        and UUID information included.
    """
    if path is None:
        uuid = build_uuid()
        kwargs["uuid"] = uuid
        kwargs["path"] = build_log_path_from_source(project, entity_type, name, uuid, source)
    else:
        kwargs["path"] = path
    return kwargs



# SPDX-FileCopyrightText: © 2025 DSLab - Fondazione Bruno Kessler
#
# SPDX-License-Identifier: Apache-2.0

from __future__ import annotations

import typing

from digitalhub.utils.exceptions import BuilderError

if typing.TYPE_CHECKING:
    from digitalhub.stores.readers.data._base.builder import ReaderBuilder
    from digitalhub.stores.readers.data._base.reader import DataframeReader


class ReaderFactory:
    """
    Reader factory class.
    """

    def __init__(self) -> None:
        self._engine_builders: dict[str, ReaderBuilder] = None
        self._dataframe_builders: dict[str, ReaderBuilder] = None
        self._default: str = None

    def add_builder(self, engine: str, dataframe: str, builder: ReaderBuilder) -> None:
        """
        Add a builder to the factory.

        Parameters
        ----------
        name : str
            Reader name.
        builder : DataframeReader
            Builder object.
        """
        if self._engine_builders is None:
            self._engine_builders = {}
        if engine in self._engine_builders:
            raise BuilderError(f"Builder for engine '{engine}' already exists.")
        self._engine_builders[engine] = builder

        if self._dataframe_builders is None:
            self._dataframe_builders = {}
        if dataframe in self._dataframe_builders:
            raise BuilderError(f"Builder for dataframe '{dataframe}' already exists.")
        self._dataframe_builders[dataframe] = builder

    def build(self, engine: str | None = None, dataframe: str | None = None, **kwargs) -> DataframeReader:
        """
        Build reader object.

        Parameters
        ----------
        engine : str | None
            Engine name.
        dataframe : str | None
            Dataframe name.
        **kwargs : dict
            Keyword arguments.

        Returns
        -------
        DataframeReader
            Reader object.
        """
        if (engine is None) == (dataframe is None):
            raise BuilderError("Either engine or dataframe must be provided.")
        if engine is not None:
            return self._engine_builders[engine].build(**kwargs)
        return self._dataframe_builders[dataframe].build(**kwargs)

    def list_supported_engines(self) -> list[str]:
        """
        List supported engines.

        Returns
        -------
        list[str]
            List of supported engines.
        """
        return list(self._engine_builders.keys())

    def list_supported_dataframes(self) -> list[str]:
        """
        List supported dataframes.

        Returns
        -------
        list[str]
            List of supported dataframes.
        """
        return list(self._dataframe_builders.keys())

    def set_default(self, engine: str) -> None:
        """
        Set default engine.

        Parameters
        ----------
        engine : str
            Engine name.
        """
        if engine not in self._engine_builders:
            raise BuilderError(f"Engine {engine} not found.")
        self._default = engine

    def get_default(self) -> str:
        """
        Get default engine.

        Returns
        -------
        str
            Default engine.
        """
        if self._default is None:
            raise BuilderError("No default engine set.")
        return self._default


factory = ReaderFactory()

try:
    from digitalhub.stores.readers.data.pandas.builder import ReaderBuilderPandas

    factory.add_builder(
        ReaderBuilderPandas.ENGINE,
        ReaderBuilderPandas.DATAFRAME_CLASS,
        ReaderBuilderPandas(),
    )
    factory.set_default(ReaderBuilderPandas.ENGINE)

except ImportError:
    pass

try:
    from digitalhub.stores.readers.data.polars.builder import ReaderBuilderPolars

    factory.add_builder(
        ReaderBuilderPolars.ENGINE,
        ReaderBuilderPolars.DATAFRAME_CLASS,
        ReaderBuilderPolars(),
    )

except ImportError:
    pass



# SPDX-FileCopyrightText: © 2025 DSLab - Fondazione Bruno Kessler
#
# SPDX-License-Identifier: Apache-2.0

from __future__ import annotations

from digitalhub.entities._commons.enums import EntityKinds
from digitalhub.entities.trigger._base.builder import TriggerBuilder
from digitalhub.entities.trigger.scheduler.entity import TriggerScheduler
from digitalhub.entities.trigger.scheduler.spec import TriggerSpecScheduler, TriggerValidatorScheduler
from digitalhub.entities.trigger.scheduler.status import TriggerStatusScheduler


class TriggerSchedulerBuilder(TriggerBuilder):
    """
    TriggerScheduler builder.
    """

    ENTITY_CLASS = TriggerScheduler
    ENTITY_SPEC_CLASS = TriggerSpecScheduler
    ENTITY_SPEC_VALIDATOR = TriggerValidatorScheduler
    ENTITY_STATUS_CLASS = TriggerStatusScheduler
    ENTITY_KIND = EntityKinds.TRIGGER_SCHEDULER.value



# SPDX-FileCopyrightText: © 2025 DSLab - Fondazione Bruno Kessler
#
# SPDX-License-Identifier: Apache-2.0

from __future__ import annotations

import typing

from digitalhub.entities._commons.enums import EntityTypes
from digitalhub.entities._processors.processors import context_processor

if typing.TYPE_CHECKING:
    from digitalhub.entities.function._base.entity import Function

ENTITY_TYPE = EntityTypes.FUNCTION.value


def new_function(
    project: str,
    name: str,
    kind: str,
    uuid: str | None = None,
    description: str | None = None,
    labels: list[str] | None = None,
    embedded: bool = False,
    **kwargs,
) -> Function:
    """
    Create a Function instance with the given parameters.

    Parameters
    ----------
    project : str
        Project name.
    name : str
        Object name.
    kind : str
        Kind the object.
    uuid : str
        ID of the object.
    description : str
        Description of the object (human readable).
    labels : list[str]
        List of labels.
    embedded : bool
        Flag to determine if object spec must be embedded in project spec.
    **kwargs : dict
        Spec keyword arguments.

    Returns
    -------
    Function
        Object instance.

    Examples
    --------
    >>> obj = new_function(project="my-project",
    >>>                    name="my-function",
    >>>                    kind="python",
    >>>                    code_src="function.py",
    >>>                    handler="function-handler")
    """
    return context_processor.create_context_entity(
        project=project,
        name=name,
        kind=kind,
        uuid=uuid,
        description=description,
        labels=labels,
        embedded=embedded,
        entity_type=ENTITY_TYPE,
        **kwargs,
    )


def get_function(
    identifier: str,
    project: str | None = None,
    entity_id: str | None = None,
) -> Function:
    """
    Get object from backend.

    Parameters
    ----------
    identifier : str
        Entity key (store://...) or entity name.
    project : str
        Project name.
    entity_id : str
        Entity ID.

    Returns
    -------
    Function
        Object instance.

    Examples
    --------
    Using entity key:
    >>> obj = get_function("store://my-function-key")

    Using entity name:
    >>> obj = get_function("my-function-name"
    >>>                    project="my-project",
    >>>                    entity_id="my-function-id")
    """
    return context_processor.read_context_entity(
        identifier=identifier,
        entity_type=ENTITY_TYPE,
        project=project,
        entity_id=entity_id,
    )


def get_function_versions(
    identifier: str,
    project: str | None = None,
) -> list[Function]:
    """
    Get object versions from backend.

    Parameters
    ----------
    identifier : str
        Entity key (store://...) or entity name.
    project : str
        Project name.

    Returns
    -------
    list[Function]
        List of object instances.

    Examples
    --------
    Using entity key:
    >>> obj = get_function_versions("store://my-function-key")

    Using entity name:
    >>> obj = get_function_versions("my-function-name"
    >>>                             project="my-project")
    """
    return context_processor.read_context_entity_versions(
        identifier=identifier,
        entity_type=ENTITY_TYPE,
        project=project,
    )


def list_functions(
    project: str,
    q: str | None = None,
    name: str | None = None,
    kind: str | None = None,
    user: str | None = None,
    state: str | None = None,
    created: str | None = None,
    updated: str | None = None,
    versions: str | None = None,
) -> list[Function]:
    """
    List all latest version objects from backend.

    Parameters
    ----------
    project : str
        Project name.
    q : str
        Query string to filter objects.
    name : str
        Object name.
    kind : str
        Kind of the object.
    user : str
        User that created the object.
    state : str
        Object state.
    created : str
        Creation date filter.
    updated : str
        Update date filter.
    versions : str
        Object version, default is latest.

    Returns
    -------
    list[Function]
        List of object instances.

    Examples
    --------
    >>> objs = list_functions(project="my-project")
    """
    return context_processor.list_context_entities(
        project=project,
        entity_type=ENTITY_TYPE,
        q=q,
        name=name,
        kind=kind,
        user=user,
        state=state,
        created=created,
        updated=updated,
        versions=versions,
    )


def import_function(
    file: str | None = None,
    key: str | None = None,
    reset_id: bool = False,
    context: str | None = None,
) -> Function:
    """
    Import an object from a YAML file or from a storage key.

    Parameters
    ----------
    file : str
        Path to the YAML file.
    key : str
        Entity key (store://...).
    reset_id : bool
        Flag to determine if the ID of executable entities should be reset.
    context : str
        Project name to use for context resolution.

    Returns
    -------
    Function
        Object instance.

    Examples
    --------
    >>> obj = import_function("my-function.yaml")
    """
    return context_processor.import_executable_entity(file, key, reset_id, context)


def load_function(file: str) -> Function:
    """
    Load object from a YAML file and update an existing object into the backend.

    Parameters
    ----------
    file : str
        Path to YAML file.

    Returns
    -------
    Function
        Object instance.

    Examples
    --------
    >>> obj = load_function("my-function.yaml")
    """
    return context_processor.load_executable_entity(file)


def update_function(entity: Function) -> Function:
    """
    Update object. Note that object spec are immutable.

    Parameters
    ----------
    entity : Function
        Object to update.

    Returns
    -------
    Function
        Entity updated.

    Examples
    --------
    >>> obj = update_function(obj)
    """
    return context_processor.update_context_entity(
        project=entity.project,
        entity_type=entity.ENTITY_TYPE,
        entity_id=entity.id,
        entity_dict=entity.to_dict(),
    )


def delete_function(
    identifier: str,
    project: str | None = None,
    entity_id: str | None = None,
    delete_all_versions: bool = False,
    cascade: bool = True,
) -> dict:
    """
    Delete object from backend.

    Parameters
    ----------
    identifier : str
        Entity key (store://...) or entity name.
    project : str
        Project name.
    entity_id : str
        Entity ID.
    delete_all_versions : bool
        Delete all versions of the named entity.
        If True, use entity name instead of entity key as identifier.
    cascade : bool
        Cascade delete.

    Returns
    -------
    dict
        Response from backend.

    Examples
    --------
    If delete_all_versions is False:
    >>> obj = delete_function("store://my-function-key")

    Otherwise:
    >>> obj = delete_function("function-name",
    >>>                       project="my-project",
    >>>                       delete_all_versions=True)
    """
    return context_processor.delete_context_entity(
        identifier=identifier,
        entity_type=ENTITY_TYPE,
        project=project,
        entity_id=entity_id,
        delete_all_versions=delete_all_versions,
        cascade=cascade,
    )



# SPDX-FileCopyrightText: © 2025 DSLab - Fondazione Bruno Kessler
#
# SPDX-License-Identifier: Apache-2.0

from __future__ import annotations

import typing

from digitalhub.entities._processors.base.crud import BaseEntityCRUDProcessor
from digitalhub.entities._processors.base.import_export import BaseEntityImportExportProcessor
from digitalhub.entities._processors.base.special_ops import BaseEntitySpecialOpsProcessor

if typing.TYPE_CHECKING:
    from digitalhub.entities.project._base.entity import Project


class BaseEntityOperationsProcessor:
    """
    Processor for base entity operations.

    This class handles CRUD operations and other entity management tasks
    for base-level entities (primarily projects). It interacts with the
    client layer to perform backend operations and manages entity lifecycle
    including creation, reading, updating, deletion, and sharing.

    Uses composition with specialized processors for different operation types.
    """

    def __init__(self):
        self.crud_processor = BaseEntityCRUDProcessor()
        self.import_export_processor = BaseEntityImportExportProcessor()
        self.special_ops_processor = BaseEntitySpecialOpsProcessor()

    ##############################
    # CRUD base entity
    ##############################

    def create_project_entity(
        self,
        _entity: Project | None = None,
        **kwargs,
    ) -> Project:
        """
        Create a project entity in the backend.

        Creates a new project either from an existing entity object or
        by building one from the provided parameters. Handles both
        local and remote backend creation.

        Parameters
        ----------
        _entity : Project
            An existing project entity object to create. If None,
            a new entity will be built from kwargs.
        **kwargs : dict
            Parameters for entity creation, including 'local' flag
            and entity-specific parameters.

        Returns
        -------
        Project
            The created project entity with backend data populated.
        """
        return self.crud_processor.create_project_entity(_entity, **kwargs)

    def read_project_entity(
        self,
        entity_type: str,
        entity_name: str,
        **kwargs,
    ) -> Project:
        """
        Read a project entity from the backend.

        Retrieves project data from the backend and constructs a
        Project entity object with the retrieved data.

        Parameters
        ----------
        entity_type : str
            The type of entity to read (typically 'project').
        entity_name : str
            The name identifier of the project to read.
        **kwargs : dict
            Additional parameters including 'local' flag and
            API call parameters.

        Returns
        -------
        Project
            The project entity object populated with backend data.
        """
        return self.crud_processor.read_project_entity(entity_type, entity_name, **kwargs)

    def update_project_entity(
        self,
        entity_type: str,
        entity_name: str,
        entity_dict: dict,
        **kwargs,
    ) -> Project:
        """
        Update a project entity in the backend.

        Updates an existing project with new data and returns the
        updated Project entity object.

        Parameters
        ----------
        entity_type : str
            The type of entity to update (typically 'project').
        entity_name : str
            The name identifier of the project to update.
        entity_dict : dict
            The updated project data dictionary.
        **kwargs : dict
            Additional parameters including 'local' flag and
            API call parameters.

        Returns
        -------
        Project
            The updated project entity object.
        """
        return self.crud_processor.update_project_entity(entity_type, entity_name, entity_dict, **kwargs)

    def delete_project_entity(
        self,
        entity_type: str,
        entity_name: str,
        **kwargs,
    ) -> dict:
        """
        Delete a project entity from the backend.

        Deletes a project from the backend and optionally cleans up
        the associated context. Handles both local and remote backends.

        Parameters
        ----------
        entity_type : str
            The type of entity to delete (typically 'project').
        entity_name : str
            The name identifier of the project to delete.
        **kwargs : dict
            Additional parameters including 'local' flag, 'clean_context'
            flag (default True), and API call parameters.

        Returns
        -------
        dict
            Response data from the backend delete operation.
        """
        return self.crud_processor.delete_project_entity(entity_type, entity_name, **kwargs)

    def list_project_entities(
        self,
        entity_type: str,
        **kwargs,
    ) -> list[Project]:
        """
        List project entities from the backend.

        Retrieves a list of projects from the backend and converts
        them to Project entity objects.

        Parameters
        ----------
        entity_type : str
            The type of entities to list (typically 'project').
        **kwargs : dict
            Additional parameters including 'local' flag and
            API call parameters for filtering or pagination.

        Returns
        -------
        list[Project]
            List of project entity objects.
        """
        return self.crud_processor.list_project_entities(entity_type, **kwargs)

    ##############################
    # Import/Export operations
    ##############################

    def import_project_entity(
        self,
        file: str,
        **kwargs,
    ) -> Project:
        """
        Import a project entity from a YAML file and create it in the backend.

        Reads project configuration from a YAML file, creates a new project
        entity in the backend, and imports any related entities defined
        in the file. Raises an error if the project already exists.

        Parameters
        ----------
        file : str
            Path to the YAML file containing project configuration.
        **kwargs : dict
            Additional parameters including 'local' and 'reset_id' flags.

        Returns
        -------
        Project
            The imported and created project entity.

        Raises
        ------
        EntityError
            If the project already exists in the backend.
        """
        return self.import_export_processor.import_project_entity(self.crud_processor, file, **kwargs)

    def load_project_entity(
        self,
        file: str,
        **kwargs,
    ) -> Project:
        """
        Load a project entity from a YAML file and update it in the backend.

        Reads project configuration from a YAML file and updates an existing
        project in the backend. If the project doesn't exist, it creates a
        new one. Also loads any related entities defined in the file.

        Parameters
        ----------
        file : str
            Path to the YAML file containing project configuration.
        **kwargs : dict
            Additional parameters including 'local' flag.

        Returns
        -------
        Project
            The loaded and updated project entity.
        """
        return self.import_export_processor.load_project_entity(self.crud_processor, file, **kwargs)

    ##############################
    # Base entity operations
    ##############################

    def build_project_key(
        self,
        entity_id: str,
    ) -> str:
        """
        Build a storage key for a project entity.

        Creates a standardized key string for project identification
        and storage, handling both local and remote client contexts.

        Parameters
        ----------
        entity_id : str
            The unique identifier of the project entity.

        Returns
        -------
        str
            The constructed project entity key string.
        """
        return self.special_ops_processor.build_project_key(entity_id)

    def share_project_entity(
        self,
        entity_type: str,
        entity_name: str,
        **kwargs,
    ) -> None:
        """
        Share or unshare a project entity with a user.

        Manages project access permissions by sharing the project with
        a specified user or removing user access. Handles both sharing
        and unsharing operations based on the 'unshare' parameter.

        Parameters
        ----------
        entity_type : str
            The type of entity to share (typically 'project').
        entity_name : str
            The name identifier of the project to share.
        **kwargs : dict
            Additional parameters including:
            - 'user': username to share with/unshare from
            - 'unshare': boolean flag for unsharing (default False)
            - 'local': boolean flag for local backend

        Raises
        ------
        ValueError
            If trying to unshare from a user who doesn't have access.
        """
        return self.special_ops_processor.share_project_entity(entity_type, entity_name, **kwargs)



# SPDX-FileCopyrightText: © 2025 DSLab - Fondazione Bruno Kessler
#
# SPDX-License-Identifier: Apache-2.0

from __future__ import annotations

import typing

from requests import request

from digitalhub.stores.client.common.config import get_client_config
from digitalhub.stores.client.http.response import ResponseProcessor
from digitalhub.utils.exceptions import BackendError, UnauthorizedError

if typing.TYPE_CHECKING:
    from digitalhub.stores.client.auth.client_configurator import ClientConfigurator


class HttpRequestHandler:
    """
    Handles HTTP request execution for DHCore client.

    Encapsulates all HTTP communication logic including request execution,
    automatic token refresh on authentication failures, and response processing.
    Works in coordination with configurator for authentication and response
    processor for parsing.
    """

    def __init__(self, configurator: ClientConfigurator) -> None:
        self._configurator = configurator
        self._response_processor = ResponseProcessor()

    def execute_request(self, method: str, api: str, **kwargs) -> dict:
        """
        Execute API call with full URL construction and authentication.

        Parameters
        ----------
        method : str
            HTTP method type (GET, POST, PUT, DELETE, etc.).
        api : str
            API endpoint path to call.
        **kwargs : dict
            Additional HTTP request arguments.

        Returns
        -------
        dict
            Response from the API call.
        """
        full_kwargs = self._configurator.get_auth_parameters(kwargs)
        url = f"{self._configurator.get_endpoint()}{api}"
        return self._execute_request(method, url, **full_kwargs)

    def _execute_request(
        self,
        method: str,
        url: str,
        **kwargs,
    ) -> dict:
        """
        Execute HTTP request with automatic handling.

        Sends HTTP request with authentication, handles token refresh on 401 errors,
        validates API version compatibility, and parses response. Uses 60-second
        timeout by default.

        Parameters
        ----------
        method : str
            HTTP method (GET, POST, PUT, DELETE, etc.).
        url : str
            Complete URL to request.
        **kwargs : dict
            Additional HTTP request arguments (headers, params, data, etc.).

        Returns
        -------
        dict
            Parsed response body as dictionary.
        """
        response = request(method, url, timeout=get_client_config().http_timeout, **kwargs)
        try:
            return self._response_processor.process(response)
        except UnauthorizedError as e:
            if self._configurator.evaluate_refresh():
                kwargs = self._configurator.get_auth_parameters(kwargs)
                return self._execute_request(method, url, **kwargs)
            raise e
        except BackendError as e:
            raise e



# SPDX-FileCopyrightText: © 2025 DSLab - Fondazione Bruno Kessler
#
# SPDX-License-Identifier: Apache-2.0

from __future__ import annotations

from digitalhub.entities._commons.enums import EntityKinds
from digitalhub.entities.trigger._base.builder import TriggerBuilder
from digitalhub.entities.trigger.automl.entity import TriggerAutoml
from digitalhub.entities.trigger.automl.spec import TriggerSpecAutoml, TriggerValidatorAutoml
from digitalhub.entities.trigger.automl.status import TriggerStatusAutoml


class TriggerAutomlBuilder(TriggerBuilder):
    """
    TriggerAutoml builder.
    """

    ENTITY_CLASS = TriggerAutoml
    ENTITY_SPEC_CLASS = TriggerSpecAutoml
    ENTITY_SPEC_VALIDATOR = TriggerValidatorAutoml
    ENTITY_STATUS_CLASS = TriggerStatusAutoml
    ENTITY_KIND = EntityKinds.TRIGGER_AUTOML.value



# SPDX-FileCopyrightText: © 2025 DSLab - Fondazione Bruno Kessler
#
# SPDX-License-Identifier: Apache-2.0

from __future__ import annotations

from digitalhub.entities._commons.enums import EntityKinds
from digitalhub.entities.model._base.builder import ModelBuilder
from digitalhub.entities.model.model.entity import ModelModel
from digitalhub.entities.model.model.spec import ModelSpecModel, ModelValidatorModel
from digitalhub.entities.model.model.status import ModelStatusModel


class ModelMlflowBuilder(ModelBuilder):
    """
    ModelModel builder.
    """

    ENTITY_CLASS = ModelModel
    ENTITY_SPEC_CLASS = ModelSpecModel
    ENTITY_SPEC_VALIDATOR = ModelValidatorModel
    ENTITY_STATUS_CLASS = ModelStatusModel
    ENTITY_KIND = EntityKinds.MODEL_MODEL.value



# SPDX-FileCopyrightText: © 2025 DSLab - Fondazione Bruno Kessler
#
# SPDX-License-Identifier: Apache-2.0

from __future__ import annotations

import typing

from digitalhub.entities.trigger._base.entity import Trigger

if typing.TYPE_CHECKING:
    from digitalhub.entities._base.entity.metadata import Metadata
    from digitalhub.entities.trigger.automl.spec import TriggerSpecAutoml
    from digitalhub.entities.trigger.automl.status import TriggerStatusAutoml


class TriggerAutoml(Trigger):
    """
    TriggerAutoml class.
    """

    def __init__(
        self,
        project: str,
        name: str,
        uuid: str,
        kind: str,
        metadata: Metadata,
        spec: TriggerSpecAutoml,
        status: TriggerStatusAutoml,
        user: str | None = None,
    ) -> None:
        super().__init__(project, name, uuid, kind, metadata, spec, status, user)

        self.spec: TriggerSpecAutoml
        self.status: TriggerStatusAutoml



# SPDX-FileCopyrightText: © 2025 DSLab - Fondazione Bruno Kessler
#
# SPDX-License-Identifier: Apache-2.0

from __future__ import annotations

from pydantic import BaseModel

_DEFAULT_FILES_STORE = "s3://datalake"


class ProfileConfig(BaseModel):
    """
    Configuration profiles.
    """

    default_files_store: str | None = _DEFAULT_FILES_STORE

    def to_dict(self) -> dict:
        return self.model_dump(by_alias=True, exclude_none=True)



# SPDX-FileCopyrightText: © 2025 DSLab - Fondazione Bruno Kessler
#
# SPDX-License-Identifier: Apache-2.0

from __future__ import annotations

from digitalhub.entities._base.entity.spec import Spec, SpecValidator


class MaterialSpec(Spec):
    """
    Material specifications.class.
    """

    def __init__(self, path: str, **kwargs) -> None:
        super().__init__(**kwargs)
        self.path = path


class MaterialValidator(SpecValidator):
    """
    Material parameters class.
    """

    path: str
    """Target path to file(s)"""



# SPDX-FileCopyrightText: © 2025 DSLab - Fondazione Bruno Kessler
#
# SPDX-License-Identifier: Apache-2.0

from __future__ import annotations

import typing

from digitalhub.entities._commons.enums import EntityKinds, EntityTypes
from digitalhub.entities._processors.processors import base_processor, context_processor
from digitalhub.entities.project.utils import setup_project
from digitalhub.utils.exceptions import BackendError

if typing.TYPE_CHECKING:
    from digitalhub.entities._base.context.entity import ContextEntity
    from digitalhub.entities.project._base.entity import Project


ENTITY_TYPE = EntityTypes.PROJECT.value


def new_project(
    name: str,
    description: str | None = None,
    labels: list[str] | None = None,
    config: dict | None = None,
    source: str | None = None,
    setup_kwargs: dict | None = None,
) -> Project:
    """
    Create a new object.

    Parameters
    ----------
    name : str
        Object name.
    description : str
        Description of the object (human readable).
    labels : list[str]
        List of labels.
    config : dict
        DHCore environment configuration.
    source : str
        The context local folder of the project.
    setup_kwargs : dict
        Setup keyword arguments passed to setup_project() function.

    Returns
    -------
    Project
        Object instance.

    Examples
    --------
    >>> obj = new_project("my-project")
    """
    if source is None:
        source = "./"
    obj = base_processor.create_project_entity(
        name=name,
        kind=EntityKinds.PROJECT_PROJECT.value,
        description=description,
        labels=labels,
        config=config,
        source=source,
    )
    return setup_project(obj, setup_kwargs)


def get_project(
    name: str,
    setup_kwargs: dict | None = None,
) -> Project:
    """
    Retrieves project details from backend.

    Parameters
    ----------
    name : str
        The Project name.
    setup_kwargs : dict
        Setup keyword arguments passed to setup_project() function.

    Returns
    -------
    Project
        Object instance.

    Examples
    --------
    >>> obj = get_project("my-project")
    """
    obj = base_processor.read_project_entity(
        entity_type=ENTITY_TYPE,
        entity_name=name,
    )
    return setup_project(obj, setup_kwargs)


def import_project(
    file: str,
    setup_kwargs: dict | None = None,
    reset_id: bool = False,
) -> Project:
    """
    Import object from a YAML file and create a new object into the backend.

    Parameters
    ----------
    file : str
        Path to YAML file.
    setup_kwargs : dict
        Setup keyword arguments passed to setup_project() function.
    reset_id : bool
        Flag to determine if the ID of project entities should be reset.

    Returns
    -------
    Project
        Object instance.

    Examples
    --------
    >>> obj = import_project("my-project.yaml")
    """
    obj = base_processor.import_project_entity(
        file=file,
        reset_id=reset_id,
    )
    return setup_project(obj, setup_kwargs)


def load_project(
    file: str,
    setup_kwargs: dict | None = None,
) -> Project:
    """
    Load object from a YAML file and update an existing object into the backend.

    Parameters
    ----------
    file : str
        Path to YAML file.
    setup_kwargs : dict
        Setup keyword arguments passed to setup_project() function.

    Returns
    -------
    Project
        Object instance.

    Examples
    --------
    >>> obj = load_project("my-project.yaml")
    """
    obj = base_processor.load_project_entity(file=file)
    return setup_project(obj, setup_kwargs)


def list_projects() -> list[Project]:
    """
    List projects in backend.


    Returns
    -------
    list
        List of objects.
    """
    return base_processor.list_project_entities(ENTITY_TYPE)


def get_or_create_project(
    name: str,
    description: str | None = None,
    labels: list[str] | None = None,
    config: dict | None = None,
    context: str | None = None,
    setup_kwargs: dict | None = None,
) -> Project:
    """
    Try to get project. If not exists, create it.

    Parameters
    ----------
    name : str
        Project name.
    config : dict
        DHCore environment configuration.
    context : str
        Folder where the project will saves its context locally.
    setup_kwargs : dict
        Setup keyword arguments passed to setup_project() function.
    **kwargs : dict
        Keyword arguments.

    Returns
    -------
    Project
        Object instance.
    """
    try:
        return get_project(
            name,
            setup_kwargs=setup_kwargs,
        )
    except BackendError:
        return new_project(
            name,
            description=description,
            labels=labels,
            config=config,
            setup_kwargs=setup_kwargs,
            source=context,
        )


def update_project(entity: Project) -> Project:
    """
    Update object. Note that object spec are immutable.

    Parameters
    ----------
    entity : Project
        Object to update.

    Returns
    -------
    Project
        The updated object.

    Examples
    --------
    >>> obj = update_project(obj)
    """
    return base_processor.update_project_entity(
        entity_type=entity.ENTITY_TYPE,
        entity_name=entity.name,
        entity_dict=entity.to_dict(),
    )


def delete_project(
    name: str,
    cascade: bool = True,
    clean_context: bool = True,
) -> dict:
    """
    Delete a project.

    Parameters
    ----------
    name : str
        Project name.
    cascade : bool
        Flag to determine if delete is cascading.
    clean_context : bool
        Flag to determine if context will be deleted.

    Returns
    -------
    dict
        Response from backend.

    Examples
    --------
    >>> delete_project("my-project")
    """
    return base_processor.delete_project_entity(
        entity_type=ENTITY_TYPE,
        entity_name=name,
        cascade=cascade,
        clean_context=clean_context,
    )


def search_entity(
    project_name: str,
    query: str | None = None,
    entity_types: list[str] | None = None,
    name: str | None = None,
    kind: str | None = None,
    created: str | None = None,
    updated: str | None = None,
    description: str | None = None,
    labels: list[str] | None = None,
) -> list[ContextEntity]:
    """
    Search objects from backend.

    Parameters
    ----------
    project_name : str
        Project name.
    query : str
        Search query.
    entity_types : list[str]
        Entity types.
    name : str
        Entity name.
    kind : str
        Entity kind.
    created : str
        Entity creation date.
    updated : str
        Entity update date.
    description : str
        Entity description.
    labels : list[str]
        Entity labels.

    Returns
    -------
    list[ContextEntity]
        List of object instances.
    """
    return context_processor.search_entity(
        project_name,
        query=query,
        entity_types=entity_types,
        name=name,
        kind=kind,
        created=created,
        updated=updated,
        description=description,
        labels=labels,
    )



# SPDX-FileCopyrightText: © 2025 DSLab - Fondazione Bruno Kessler
#
# SPDX-License-Identifier: Apache-2.0

from __future__ import annotations

from abc import ABC, abstractmethod


class ParameterStrategy(ABC):
    """
    Abstract base class for parameter building strategies.

    Defines interface for parameter builders that handle specific
    operations or parameter types.
    """

    @abstractmethod
    def build(self, **kwargs) -> dict:
        """
        Build parameters for specific operation.

        Parameters
        ----------
        **kwargs : dict
            Raw parameters to transform.

        Returns
        -------
        dict
            Formatted parameters.
        """
        pass


class SearchParameterStrategy(ParameterStrategy):
    """
    Strategy for building Solr search parameters.

    Handles complex search filter construction including entity types,
    names, kinds, time ranges, descriptions, and labels.
    """

    def build(self, **kwargs) -> dict:
        """
        Build search parameters with Solr filters.

        Parameters
        ----------
        **kwargs : dict
            Raw search parameters including entity_types, name, kind,
            created, updated, description, labels, etc.

        Returns
        -------
        dict
            Formatted parameters with 'fq' (filter query) list.
        """
        fq = []

        # Entity types
        if (entity_types := kwargs.pop("entity_types", None)) is not None:
            entity_types = self._format_entity_types(entity_types)
            fq.append(f"type:({entity_types})")

        # Name
        if (name := kwargs.pop("name", None)) is not None:
            fq.append(f'metadata.name:"{name}"')

        # Kind
        if (kind := kwargs.pop("kind", None)) is not None:
            fq.append(f'kind:"{kind}"')

        # Time range
        created = kwargs.pop("created", None)
        updated = kwargs.pop("updated", None)
        created = created if created is not None else "*"
        updated = updated if updated is not None else "*"
        fq.append(f"metadata.updated:[{created} TO {updated}]")

        # Description
        if (description := kwargs.pop("description", None)) is not None:
            fq.append(f'metadata.description:"{description}"')

        # Labels
        if (labels := kwargs.pop("labels", None)) is not None:
            labels = self._format_labels(labels)
            fq.append(f"metadata.labels:({labels})")

        kwargs["fq"] = fq
        return kwargs

    @staticmethod
    def _format_entity_types(entity_types: list[str] | str) -> str:
        """
        Format entity types for Solr query.

        Parameters
        ----------
        entity_types : list[str] or str
            Entity type or types to format.

        Returns
        -------
        str
            Formatted entity types string.
        """
        if not isinstance(entity_types, list):
            entity_types = [entity_types]
        if len(entity_types) == 1:
            return entity_types[0]
        return " OR ".join(entity_types)

    @staticmethod
    def _format_labels(labels: list[str]) -> str:
        """
        Format labels for Solr query.

        Parameters
        ----------
        labels : list[str]
            Labels to format.

        Returns
        -------
        str
            Formatted labels string.
        """
        if len(labels) == 1:
            return labels[0]
        return " AND ".join(labels)


class ListParameterStrategy(ParameterStrategy):
    """
    Strategy for building list operation parameters.

    Handles filtering parameters for entity list operations including
    query, name, kind, user, state, timestamps, and relationships.
    """

    ALLOWED_PARAMS = [
        "q",
        "name",
        "kind",
        "user",
        "state",
        "created",
        "updated",
        "versions",
        "function",
        "workflow",
        "action",
        "task",
    ]

    def build(self, **kwargs) -> dict:
        """
        Build list operation parameters.

        Parameters
        ----------
        **kwargs : dict
            Raw list parameters.

        Returns
        -------
        dict
            Formatted parameters with allowed list filters.
        """
        # Extract only allowed parameters
        list_params = {k: kwargs.get(k, None) for k in self.ALLOWED_PARAMS}
        list_params = {k: v for k, v in list_params.items() if v is not None}

        # Remove from kwargs and return filtered params
        for k in self.ALLOWED_PARAMS:
            kwargs.pop(k, None)

        kwargs["list_params"] = list_params
        return kwargs



# SPDX-FileCopyrightText: © 2025 DSLab - Fondazione Bruno Kessler
#
# SPDX-License-Identifier: Apache-2.0

from __future__ import annotations

import typing

from digitalhub.entities._base.material.entity import MaterialEntity
from digitalhub.entities._base.metrics.entity import MetricsEntity
from digitalhub.entities._commons.enums import EntityTypes

if typing.TYPE_CHECKING:
    from digitalhub.entities._base.entity.metadata import Metadata
    from digitalhub.entities.model._base.spec import ModelSpec
    from digitalhub.entities.model._base.status import ModelStatus


class Model(MaterialEntity, MetricsEntity):
    """
    A class representing a model.
    """

    ENTITY_TYPE = EntityTypes.MODEL.value

    def __init__(
        self,
        project: str,
        name: str,
        uuid: str,
        kind: str,
        metadata: Metadata,
        spec: ModelSpec,
        status: ModelStatus,
        user: str | None = None,
    ) -> None:
        super().__init__(project, name, uuid, kind, metadata, spec, status, user)
        self.spec: ModelSpec
        self.status: ModelStatus



# SPDX-FileCopyrightText: © 2025 DSLab - Fondazione Bruno Kessler
#
# SPDX-License-Identifier: Apache-2.0

from __future__ import annotations

import typing

from digitalhub.context.api import get_context
from digitalhub.entities._commons.utils import get_project_from_key, is_valid_key, parse_entity_key
from digitalhub.utils.exceptions import EntityError

if typing.TYPE_CHECKING:
    from digitalhub.context.context import Context


def parse_identifier(
    identifier: str,
    project: str | None = None,
    entity_type: str | None = None,
    entity_kind: str | None = None,
    entity_id: str | None = None,
) -> tuple[str, str, str | None, str | None, str | None]:
    """
    Parse and validate entity identifier into its components.

    Processes an entity identifier that can be either a full entity key
    (store://) or a simple entity name. When using a simple name,
    additional parameters must be provided for proper identification.

    Parameters
    ----------
    identifier : str
        The entity identifier to parse. Can be either a full entity key
        (store://project/entity_type/kind/name:id) or a simple entity name.
    project : str
        The project name. Required when identifier is not a full key.
    entity_type : str
        The entity type. Required when identifier is not a full key.
    entity_kind : str
        The entity kind specification.
    entity_id : str
        The entity version identifier.

    Returns
    -------
    tuple[str, str, str | None, str | None, str | None]
        A tuple containing (project_name, entity_type, entity_kind,
        entity_name, entity_id) parsed from the identifier.

    Raises
    ------
    ValueError
        If identifier is not a full key and project or entity_type is None.
    """
    if not is_valid_key(identifier):
        if project is None or entity_type is None:
            raise ValueError("Project and entity type must be specified.")
        return project, entity_type, entity_kind, identifier, entity_id
    return parse_entity_key(identifier)


def get_context_from_identifier(
    identifier: str,
    project: str | None = None,
) -> Context:
    """
    Retrieve context instance from entity identifier or project name.

    Extracts project information from the identifier and returns the
    corresponding context. If the identifier is not a full key, the
    project parameter must be provided explicitly.

    Parameters
    ----------
    identifier : str
        The entity identifier to extract context from. Can be either
        a full entity key (store://...) or a simple entity name.
    project : str
        The project name. Required when identifier is not a full key.

    Returns
    -------
    Context
        The context instance associated with the identified project.

    Raises
    ------
    EntityError
        If identifier is not a full key and project parameter is None.
    """
    if not is_valid_key(identifier):
        if project is None:
            raise EntityError("Specify project if you do not specify entity key.")
    else:
        project = get_project_from_key(identifier)

    return get_context(project)



# SPDX-FileCopyrightText: © 2025 DSLab - Fondazione Bruno Kessler
#
# SPDX-License-Identifier: Apache-2.0

from __future__ import annotations

import typing

from digitalhub.factory.registry import registry
from digitalhub.utils.exceptions import BuilderError

if typing.TYPE_CHECKING:
    from digitalhub.entities._base.entity.entity import Entity
    from digitalhub.entities._base.entity.metadata import Metadata
    from digitalhub.entities._base.entity.spec import Spec, SpecValidator
    from digitalhub.entities._base.entity.status import Status


class EntityFactory:
    """
    Factory for creating and managing entity builders.

    This class handles the creation of entities and their components
    through their respective builders, using a centralized registry.
    """

    def _call_builder_method(self, kind: str, method_name: str, *args, **kwargs):
        """
        Helper method to get a builder and call a method on it.

        Parameters
        ----------
        kind : str
            The kind of builder to retrieve.
        method_name : str
            The name of the method to call on the builder.
        *args
            Positional arguments to pass to the method.
        **kwargs
            Keyword arguments to pass to the method.

        Returns
        -------
        Any
            The result of calling the method on the builder.
        """
        builder = registry.get_entity_builder(kind)
        return getattr(builder, method_name)(*args, **kwargs)

    def build_entity_from_params(self, **kwargs) -> Entity:
        """
        Build an entity from parameters.

        Parameters
        ----------
        **kwargs
            Entity parameters.

        Returns
        -------
        Entity
            Entity object.
        """
        kind = self._get_kind(**kwargs)
        builder = registry.get_entity_builder(kind)
        return builder.build(**kwargs)

    def build_entity_from_dict(self, obj: dict) -> Entity:
        """
        Build an entity from a dictionary.

        Parameters
        ----------
        obj : dict
            Dictionary with entity data.

        Returns
        -------
        Entity
            Entity object.
        """
        kind = self._get_kind(**obj)
        builder = registry.get_entity_builder(kind)
        return builder.from_dict(obj)

    def build_spec(self, kind_to_build_from: str, **kwargs) -> Spec:
        """
        Build an entity spec.

        Parameters
        ----------
        kind_to_build_from : str
            Entity type.
        **kwargs
            Additional spec parameters.

        Returns
        -------
        Spec
            Spec object.
        """
        return self._call_builder_method(kind_to_build_from, "build_spec", **kwargs)

    def build_metadata(self, kind_to_build_from: str, **kwargs) -> Metadata:
        """
        Build an entity metadata.

        Parameters
        ----------
        kind_to_build_from : str
            Entity type.
        **kwargs
            Additional metadata parameters.

        Returns
        -------
        Metadata
            Metadata object.
        """
        return self._call_builder_method(kind_to_build_from, "build_metadata", **kwargs)

    def build_status(self, kind_to_build_from: str, **kwargs) -> Status:
        """
        Build an entity status.

        Parameters
        ----------
        kind_to_build_from : str
            Entity type.
        **kwargs
            Additional status parameters.

        Returns
        -------
        Status
            Status object.
        """
        return self._call_builder_method(kind_to_build_from, "build_status", **kwargs)

    def get_entity_type_from_kind(self, kind: str) -> str:
        """
        Get entity type from builder.

        Parameters
        ----------
        kind : str
            Entity type.

        Returns
        -------
        str
            Entity type.
        """
        return self._call_builder_method(kind, "get_entity_type")

    def get_executable_kind(self, kind: str) -> str:
        """
        Get executable kind.

        Parameters
        ----------
        kind : str
            Kind.

        Returns
        -------
        str
            Executable kind.
        """
        return self._call_builder_method(kind, "get_executable_kind")

    def get_action_from_task_kind(self, kind: str, task_kind: str) -> str:
        """
        Get action from task.

        Parameters
        ----------
        kind : str
            Kind.
        task_kind : str
            Task kind.

        Returns
        -------
        str
            Action.
        """
        return self._call_builder_method(kind, "get_action_from_task_kind", task_kind)

    def get_task_kind_from_action(self, kind: str, action: str) -> str:
        """
        Get task kind from action.

        Parameters
        ----------
        kind : str
            Kind.
        action : str
            Action.

        Returns
        -------
        str
            Task kind.
        """
        return self._call_builder_method(kind, "get_task_kind_from_action", action)

    def get_run_kind_from_action(self, kind: str, action: str) -> str:
        """
        Get run kind.

        Parameters
        ----------
        kind : str
            Kind.

        Returns
        -------
        str
            Run kind.
        """
        return self._call_builder_method(kind, "get_run_kind_from_action", action)

    def get_all_kinds(self, kind: str) -> list[str]:
        """
        Get all kinds.

        Parameters
        ----------
        kind : str
            Kind.

        Returns
        -------
        list of str
            All kinds.
        """
        return self._call_builder_method(kind, "get_all_kinds")

    def get_spec_validator(self, kind: str) -> SpecValidator:
        """
        Get spec validators.

        Parameters
        ----------
        kind : str
            Kind.

        Returns
        -------
        SpecValidator
            Spec validator.
        """
        return self._call_builder_method(kind, "get_spec_validator")

    @staticmethod
    def _get_kind(**kwargs) -> str:
        """
        Extract the 'kind' from parameters.

        Parameters
        ----------
        **kwargs
            Entity parameters.

        Returns
        -------
        str
            The kind of the entity.

        Raises
        ------
        BuilderError
            If 'kind' is not found in parameters.
        """
        try:
            return kwargs["kind"]
        except KeyError:
            raise BuilderError("Missing 'kind' parameter.")


# Global instance
entity_factory = EntityFactory()



# SPDX-FileCopyrightText: © 2025 DSLab - Fondazione Bruno Kessler
#
# SPDX-License-Identifier: Apache-2.0

from __future__ import annotations

from typing import Any

from digitalhub.stores.client.common.config import get_client_config
from digitalhub.utils.exceptions import ClientError
from digitalhub.utils.uri_utils import has_remote_scheme


def ensure_headers(**kwargs) -> dict:
    """
    Initialize headers dictionary in kwargs.

    Ensures parameter dictionary has 'headers' key for HTTP headers,
    guaranteeing consistent structure for all parameter building methods.

    Parameters
    ----------
    **kwargs : dict
        Keyword arguments to format. May be empty or contain various
        parameters for API operations.

    Returns
    -------
    dict
        Dictionary with guaranteed 'headers' key containing
        empty dict if not already present.
    """
    if "headers" not in kwargs:
        kwargs["headers"] = {}
    return kwargs


def ensure_params(**kwargs) -> dict:
    """
    Initialize parameter dictionary with query parameters structure.

    Ensures parameter dictionary has 'params' key for HTTP query parameters,
    guaranteeing consistent structure for all parameter building methods.

    Parameters
    ----------
    **kwargs : dict
        Keyword arguments to format. May be empty or contain various
        parameters for API operations.

    Returns
    -------
    dict
        Parameters dictionary with guaranteed 'params' key containing
        empty dict if not already present.
    """
    if "params" not in kwargs:
        kwargs["params"] = {}
    return kwargs


def add_param(key: str, value: Any | None, **kwargs) -> dict:
    """
    Add a single query parameter to kwargs.

    Parameters
    ----------
    key : str
        Parameter key.
    value : Any
        Parameter value.
    **kwargs : dict
        Keyword arguments to format. May be empty or contain various
        parameters for API operations.

    Returns
    -------
    dict
        Parameters dictionary with added key-value pair in 'params'.
    """
    kwargs["params"][key] = value
    return kwargs


def set_json_content_type(**kwargs) -> dict:
    """
    Set Content-Type header to application/json.

    Ensures that the 'Content-Type' header is set to 'application/json'
    for requests that require JSON payloads.

    Parameters
    ----------
    **kwargs : dict
        Keyword arguments to format. May be empty or contain various
        parameters for API operations.

    Returns
    -------
    dict
        Dictionary with 'Content-Type' header set to 'application/json'.
    """
    kwargs = ensure_headers(**kwargs)
    kwargs["headers"]["Content-Type"] = "application/json"
    return kwargs


def set_urlencoded_content_type(**kwargs) -> dict:
    """
    Set Content-Type header to application/x-www-form-urlencoded.

    Ensures that the 'Content-Type' header is set to
    'application/x-www-form-urlencoded' for requests that require
    form-encoded payloads.

    Parameters
    ----------
    **kwargs : dict
        Keyword arguments to format. May be empty or contain various
        parameters for API operations.

    Returns
    -------
    dict
        Dictionary with 'Content-Type' header set to
        'application/x-www-form-urlencoded'.
    """
    kwargs = ensure_headers(**kwargs)
    kwargs["headers"]["Content-Type"] = "application/x-www-form-urlencoded"
    return kwargs


def set_pagination(partial: bool = False, **kwargs) -> dict:
    """
    Ensure pagination parameters are set in kwargs.

    Parameters
    ----------
    **kwargs : dict
        Keyword arguments to format. May be empty or contain various
        parameters for API operations.

    Returns
    -------
    dict
        Pagination parameters set in 'params' of kwargs.
    """
    kwargs = ensure_params(**kwargs)

    if "page" not in kwargs["params"]:
        kwargs["params"]["page"] = get_client_config().default_page_start

    if partial:
        return kwargs

    if "size" not in kwargs["params"]:
        kwargs["params"]["size"] = get_client_config().default_page_size

    if "sort" not in kwargs["params"]:
        kwargs["params"]["sort"] = get_client_config().default_sort

    return kwargs


def read_page_number(**kwargs) -> int:
    """
    Read current page number from kwargs.

    Parameters
    ----------
    **kwargs : dict
        Keyword arguments to format. May be empty or contain various
        parameters for API operations.

    Returns
    -------
    int
        Current page number.
    """
    return kwargs["params"]["page"]


def increment_page_number(**kwargs) -> dict:
    """
    Increment page number in kwargs.

    Parameters
    ----------
    **kwargs : dict
        Keyword arguments to format. May be empty or contain various
        parameters for API operations.

    Returns
    -------
    dict
        Parameters dictionary with incremented 'page' number in 'params'.
    """
    kwargs["params"]["page"] += 1
    return kwargs


def sanitize_endpoint(endpoint: str | None = None) -> str | None:
    """
    Validate and normalize endpoint URL.

    Ensures proper HTTP/HTTPS scheme, trims whitespace, and removes trailing slashes.

    Parameters
    ----------
    endpoint : str
        Endpoint URL to sanitize.

    Returns
    -------
    str or None
        Sanitized URL or None if input was None.

    Raises
    ------
    ClientError
        If endpoint lacks http:// or https:// scheme.
    """
    if endpoint is None:
        return
    if not has_remote_scheme(endpoint):
        raise ClientError("Invalid endpoint scheme. Must start with http:// or https://.")

    endpoint = endpoint.strip()
    return endpoint.removesuffix("/")



# SPDX-FileCopyrightText: © 2025 DSLab - Fondazione Bruno Kessler
#
# SPDX-License-Identifier: Apache-2.0

from __future__ import annotations

from digitalhub.stores.client.auth.enums import ConfigurationVars, CredentialsVars
from digitalhub.stores.client.base.factory import get_client


class SqlStoreConfigurator:
    """
    Configurator class for SQL store configuration and credentials management.
    """

    def __init__(self):
        self._validate()

    def get_sql_conn_string(self) -> str:
        """
        Generate PostgreSQL connection string from stored credentials.

        Returns
        -------
        str
            A PostgreSQL connection string in the format:
            'postgresql://username:password@host:port/database'
        """
        creds = self.get_sql_credentials()
        user = creds[CredentialsVars.DB_USERNAME.value]
        password = creds[CredentialsVars.DB_PASSWORD.value]
        host = creds[ConfigurationVars.DB_HOST.value]
        port = creds[ConfigurationVars.DB_PORT.value]
        database = creds[ConfigurationVars.DB_DATABASE.value]
        return f"postgresql://{user}:{password}@{host}:{port}/{database}"

    def get_sql_credentials(self) -> dict:
        """
        Get all configured database credentials as a dictionary.

        Retrieves all available database credentials from the configured
        source and returns them as a dictionary with all credential keys
        from self.keys mapped to their values.

        Returns
        -------
        dict
            Dictionary containing all credential key-value pairs from self.keys.
            Keys correspond to database connection parameters such as
            username, password, host, port, database, and platform.
        """
        keys = [
            CredentialsVars.DB_USERNAME.value,
            CredentialsVars.DB_PASSWORD.value,
            ConfigurationVars.DB_HOST.value,
            ConfigurationVars.DB_PORT.value,
            ConfigurationVars.DB_DATABASE.value,
        ]
        creds = get_client().get_credentials_and_config()
        return {key: creds.get(key) for key in keys}

    def eval_retry(self) -> bool:
        """
        Evaluate the status of retry lifecycle.

        Returns
        -------
        bool
            True if a retry should be attempted, False otherwise.
        """
        return get_client().eval_retry()

    def _validate(self) -> None:
        """
        Validate if all required keys are present in the configuration.
        """
        required_keys = [
            ConfigurationVars.DB_HOST.value,
            ConfigurationVars.DB_PORT.value,
            ConfigurationVars.DB_DATABASE.value,
            CredentialsVars.DB_USERNAME.value,
            CredentialsVars.DB_PASSWORD.value,
        ]
        current_keys = get_client().get_credentials_and_config()
        missing_keys = []
        for key in required_keys:
            if key not in current_keys or current_keys[key] is None:
                missing_keys.append(key)
        if missing_keys:
            raise ValueError(f"Missing required variables for SQL store: {', '.join(missing_keys)}")



# SPDX-FileCopyrightText: © 2025 DSLab - Fondazione Bruno Kessler
#
# SPDX-License-Identifier: Apache-2.0

from __future__ import annotations

import typing

from digitalhub.context.api import delete_context
from digitalhub.factory.entity import entity_factory
from digitalhub.stores.client.base.factory import get_client
from digitalhub.stores.client.common.enums import ApiCategories, BackendOperations

if typing.TYPE_CHECKING:
    from digitalhub.entities.project._base.entity import Project
    from digitalhub.stores.client.base.client import Client


class BaseEntityCRUDProcessor:
    """
    Processor for core CRUD operations on base entities.

    Handles creation, reading, updating, deletion, and listing of
    base-level entities (primarily projects) within the backend.
    """

    def _create_base_entity(
        self,
        client: Client,
        entity_type: str,
        entity_dict: dict,
        **kwargs,
    ) -> dict:
        """
        Create a base entity in the backend.

        Builds the appropriate API endpoint and sends a create request
        to the backend for base-level entities.

        Parameters
        ----------
        client : Client
            The client instance to use for the API call.
        entity_type : str
            The type of entity to create (e.g., 'project').
        entity_dict : dict
            The entity data dictionary to create.
        **kwargs : dict
            Additional parameters to pass to the API call.

        Returns
        -------
        dict
            The created entity data returned from the backend.
        """
        api = client.build_api(
            ApiCategories.BASE.value,
            BackendOperations.CREATE.value,
            entity_type=entity_type,
        )
        return client.create_object(api, entity_dict, **kwargs)

    def create_project_entity(
        self,
        _entity: Project | None = None,
        **kwargs,
    ) -> Project:
        """
        Create a project entity in the backend.

        Creates a new project either from an existing entity object or
        by building one from the provided parameters.

        Parameters
        ----------
        _entity : Project
            An existing project entity object to create. If None,
            a new entity will be built from kwargs.
        **kwargs : dict
            Parameters for entity creation.

        Returns
        -------
        Project
            The created project entity with backend data populated.
        """
        if _entity is not None:
            client = _entity._client
            obj = _entity
        else:
            client = get_client()
            obj = entity_factory.build_entity_from_params(**kwargs)
        ent = self._create_base_entity(client, obj.ENTITY_TYPE, obj.to_dict())
        return entity_factory.build_entity_from_dict(ent)

    def _read_base_entity(
        self,
        client: Client,
        entity_type: str,
        entity_name: str,
        **kwargs,
    ) -> dict:
        """
        Read a base entity from the backend.

        Builds the appropriate API endpoint and sends a read request
        to retrieve entity data from the backend.

        Parameters
        ----------
        client : Client
            The client instance to use for the API call.
        entity_type : str
            The type of entity to read (e.g., 'project').
        entity_name : str
            The name identifier of the entity to read.
        **kwargs : dict
            Additional parameters to pass to the API call.

        Returns
        -------
        dict
            The entity data retrieved from the backend.
        """
        api = client.build_api(
            ApiCategories.BASE.value,
            BackendOperations.READ.value,
            entity_type=entity_type,
            entity_name=entity_name,
        )
        return client.read_object(api, **kwargs)

    def read_project_entity(
        self,
        entity_type: str,
        entity_name: str,
        **kwargs,
    ) -> Project:
        """
        Read a project entity from the backend.

        Retrieves project data from the backend and constructs a
        Project entity object with the retrieved data.

        Parameters
        ----------
        entity_type : str
            The type of entity to read (typically 'project').
        entity_name : str
            The name identifier of the project to read.
        **kwargs : dict
            Additional parameters including 'local' flag and
            API call parameters.

        Returns
        -------
        Project
            The project entity object populated with backend data.
        """
        client = get_client()
        obj = self._read_base_entity(client, entity_type, entity_name, **kwargs)
        return entity_factory.build_entity_from_dict(obj)

    def _list_base_entities(
        self,
        client: Client,
        entity_type: str,
        **kwargs,
    ) -> list[dict]:
        """
        List base entities from the backend.

        Builds the appropriate API endpoint and sends a list request
        to retrieve multiple entities from the backend.

        Parameters
        ----------
        client : Client
            The client instance to use for the API call.
        entity_type : str
            The type of entities to list (e.g., 'project').
        **kwargs : dict
            Additional parameters to pass to the API call for filtering
            or pagination.

        Returns
        -------
        list[dict]
            List of entity data dictionaries from the backend.
        """
        api = client.build_api(
            ApiCategories.BASE.value,
            BackendOperations.LIST.value,
            entity_type=entity_type,
        )
        return client.list_objects(api, **kwargs)

    def list_project_entities(
        self,
        entity_type: str,
        **kwargs,
    ) -> list[Project]:
        """
        List project entities from the backend.

        Retrieves a list of projects from the backend and converts
        them to Project entity objects.

        Parameters
        ----------
        entity_type : str
            The type of entities to list (typically 'project').
        **kwargs : dict
            Additional parameters.

        Returns
        -------
        list[Project]
            List of project entity objects.
        """
        client = get_client()
        objs = self._list_base_entities(client, entity_type, **kwargs)
        entities = []
        for obj in objs:
            ent = entity_factory.build_entity_from_dict(obj)
            entities.append(ent)
        return entities

    def _update_base_entity(
        self,
        client: Client,
        entity_type: str,
        entity_name: str,
        entity_dict: dict,
        **kwargs,
    ) -> dict:
        """
        Update a base entity in the backend.

        Builds the appropriate API endpoint and sends an update request
        to modify an existing entity in the backend.

        Parameters
        ----------
        client : Client
            The client instance to use for the API call.
        entity_type : str
            The type of entity to update (e.g., 'project').
        entity_name : str
            The name identifier of the entity to update.
        entity_dict : dict
            The updated entity data dictionary.
        **kwargs : dict
            Additional parameters to pass to the API call.

        Returns
        -------
        dict
            The updated entity data returned from the backend.
        """
        api = client.build_api(
            ApiCategories.BASE.value,
            BackendOperations.UPDATE.value,
            entity_type=entity_type,
            entity_name=entity_name,
        )
        return client.update_object(api, entity_dict, **kwargs)

    def update_project_entity(
        self,
        entity_type: str,
        entity_name: str,
        entity_dict: dict,
        **kwargs,
    ) -> Project:
        """
        Update a project entity in the backend.

        Updates an existing project with new data and returns the
        updated Project entity object.

        Parameters
        ----------
        entity_type : str
            The type of entity to update (typically 'project').
        entity_name : str
            The name identifier of the project to update.
        entity_dict : dict
            The updated project data dictionary.
        **kwargs : dict
            Additional parameters.

        Returns
        -------
        Project
            The updated project entity object.
        """
        client = get_client()
        obj = self._update_base_entity(client, entity_type, entity_name, entity_dict, **kwargs)
        return entity_factory.build_entity_from_dict(obj)

    def _delete_base_entity(
        self,
        client: Client,
        entity_type: str,
        entity_name: str,
        **kwargs,
    ) -> dict:
        """
        Delete a base entity from the backend.

        Builds the appropriate API endpoint and parameters, then sends
        a delete request to remove the entity from the backend.

        Parameters
        ----------
        client : Client
            The client instance to use for the API call.
        entity_type : str
            The type of entity to delete (e.g., 'project').
        entity_name : str
            The name identifier of the entity to delete.
        **kwargs : dict
            Additional parameters to pass to the API call, such as
            cascade deletion options.

        Returns
        -------
        dict
            Response data from the backend delete operation.
        """
        kwargs = client.build_parameters(
            ApiCategories.BASE.value,
            BackendOperations.DELETE.value,
            **kwargs,
        )
        api = client.build_api(
            ApiCategories.BASE.value,
            BackendOperations.DELETE.value,
            entity_type=entity_type,
            entity_name=entity_name,
        )
        return client.delete_object(api, **kwargs)

    def delete_project_entity(
        self,
        entity_type: str,
        entity_name: str,
        **kwargs,
    ) -> dict:
        """
        Delete a project entity from the backend.

        Deletes a project from the backend and optionally cleans up
        the associated context.

        Parameters
        ----------
        crud_processor : BaseEntityCRUDProcessor
            The CRUD processor instance for entity operations.
        entity_type : str
            The type of entity to delete (typically 'project').
        entity_name : str
            The name identifier of the project to delete.
        **kwargs : dict
            Additional parameters including 'clean_context'.

        Returns
        -------
        dict
            Response data from the backend delete operation.
        """
        if kwargs.pop("clean_context", True):
            delete_context(entity_name)
        client = get_client()
        return self._delete_base_entity(
            client,
            entity_type,
            entity_name,
            **kwargs,
        )



# SPDX-FileCopyrightText: © 2025 DSLab - Fondazione Bruno Kessler
#
# SPDX-License-Identifier: Apache-2.0

from __future__ import annotations

from pydantic import Field

from digitalhub.entities.model._base.spec import ModelSpec, ModelValidator
from digitalhub.entities.model.mlflow.models import Dataset, Signature


class ModelSpecMlflow(ModelSpec):
    """
    ModelSpecMlflow specifications.
    """

    def __init__(
        self,
        path: str,
        framework: str | None = None,
        algorithm: str | None = None,
        parameters: dict | None = None,
        flavor: str | None = None,
        model_config: dict | None = None,
        input_datasets: list[Dataset] | None = None,
        signature: Signature | None = None,
    ) -> None:
        super().__init__(path, framework, algorithm, parameters)
        self.flavor = flavor
        self.model_config = model_config
        self.input_datasets = input_datasets
        self.signature = signature


class ModelValidatorMlflow(ModelValidator):
    """
    ModelValidatorMlflow validator.
    """

    flavor: str | None = None
    """Mlflow model flavor."""

    placeholder_cfg_: dict = Field(default=None, alias="model_config")
    """Mlflow model config."""

    input_datasets: list[Dataset] | None = None
    """Mlflow input datasets."""

    signature: Signature | None = None
    """Mlflow model signature."""



# SPDX-FileCopyrightText: © 2025 DSLab - Fondazione Bruno Kessler
#
# SPDX-License-Identifier: Apache-2.0

from __future__ import annotations

from digitalhub.entities._commons.enums import EntityKinds
from digitalhub.entities.dataitem._base.builder import DataitemBuilder
from digitalhub.entities.dataitem.dataitem.entity import DataitemDataitem
from digitalhub.entities.dataitem.dataitem.spec import DataitemSpecDataitem, DataitemValidatorDataitem
from digitalhub.entities.dataitem.dataitem.status import DataitemStatusDataitem


class DataitemDataitemBuilder(DataitemBuilder):
    """
    DataitemDataitem builder.
    """

    ENTITY_CLASS = DataitemDataitem
    ENTITY_SPEC_CLASS = DataitemSpecDataitem
    ENTITY_SPEC_VALIDATOR = DataitemValidatorDataitem
    ENTITY_STATUS_CLASS = DataitemStatusDataitem
    ENTITY_KIND = EntityKinds.DATAITEM_DATAITEM.value



# SPDX-FileCopyrightText: © 2025 DSLab - Fondazione Bruno Kessler
#
# SPDX-License-Identifier: Apache-2.0

from __future__ import annotations

from pathlib import Path

import yaml

##############################
# Writers
##############################


def write_yaml(filepath: str | Path, obj: dict | list[dict]) -> None:
    """
    Write a dict or a list of dicts to a YAML file.

    Parameters
    ----------
    filepath : str or Path
        The YAML file path to write.
    obj : dict or list of dict
        The dict or list of dicts to write.
    """
    if isinstance(obj, list):
        with open(filepath, "w", encoding="utf-8") as out_file:
            yaml.dump_all(obj, out_file, sort_keys=False, default_flow_style=False)
    else:
        with open(filepath, "w", encoding="utf-8") as out_file:
            yaml.dump(obj, out_file, sort_keys=False)


def write_text(filepath: Path, text: str) -> None:
    """
    Write text to a file.

    Parameters
    ----------
    filepath : Path
        The file path to write.
    text : str
        The text to write.
    """
    filepath.write_text(text, encoding="utf-8")


##############################
# Readers
##############################


class NoDatesSafeLoader(yaml.SafeLoader):
    """
    Loader implementation to exclude implicit resolvers for YAML timestamps.

    Taken from https://stackoverflow.com/a/37958106
    """

    @classmethod
    def remove_implicit_resolver(cls, tag_to_remove: str) -> None:
        """
        Remove implicit resolvers for a particular tag.

        Takes care not to modify resolvers in super classes.
        We want to load datetimes as strings, not dates, because we
        go on to serialize as JSON which doesn't have the advanced types
        of YAML, and leads to incompatibilities down the track.

        Parameters
        ----------
        tag_to_remove : str
            The tag to remove.
        """
        if "yaml_implicit_resolvers" not in cls.__dict__:
            cls.yaml_implicit_resolvers = cls.yaml_implicit_resolvers.copy()

        for first_letter, mappings in cls.yaml_implicit_resolvers.items():
            cls.yaml_implicit_resolvers[first_letter] = [
                (tag, regexp) for tag, regexp in mappings if tag != tag_to_remove
            ]


NoDatesSafeLoader.remove_implicit_resolver("tag:yaml.org,2002:timestamp")


def read_yaml(filepath: str | Path) -> dict | list[dict]:
    """
    Read a YAML file and return its content as a dict or a list of dicts.

    Parameters
    ----------
    filepath : str or Path
        The YAML file path to read.

    Returns
    -------
    dict or list of dict
        The YAML file content.
    """
    try:
        with open(filepath, "r", encoding="utf-8") as in_file:
            data = yaml.load(in_file, Loader=NoDatesSafeLoader)

    # If yaml contains multiple documents
    except yaml.composer.ComposerError:
        with open(filepath, "r", encoding="utf-8") as in_file:
            data = list(yaml.load_all(in_file, Loader=NoDatesSafeLoader))
    return data


def read_text(filepath: str | Path) -> str:
    """
    Read a file and return its text content.

    Parameters
    ----------
    filepath : str or Path
        The file path to read.

    Returns
    -------
    str
        The file content as a string.
    """
    return Path(filepath).read_text(encoding="utf-8")



# SPDX-FileCopyrightText: © 2025 DSLab - Fondazione Bruno Kessler
#
# SPDX-License-Identifier: Apache-2.0

from __future__ import annotations

import typing

from digitalhub.stores.client.auth.enums import CredentialsVars
from digitalhub.stores.client.common.enums import AuthType
from digitalhub.stores.client.common.utils import ensure_headers

if typing.TYPE_CHECKING:
    from digitalhub.stores.client.auth.config_manager import ConfigManager


class AuthenticationHandler:
    """
    Handles authentication type evaluation and parameter injection.

    Determines the appropriate authentication method from available credentials
    and adds authentication headers/parameters to HTTP requests. Supports
    multiple authentication methods: EXCHANGE (personal access token), OAUTH2
    (access + refresh tokens), ACCESS_TOKEN (access only), and BASIC
    (username + password).
    """

    def __init__(self, credential_manager: ConfigManager) -> None:
        """
        Initialize authentication handler.

        Parameters
        ----------
        credential_manager : CredentialManager
            Credential manager instance for accessing credentials.
        """
        self._credential_manager = credential_manager
        self._auth_type: str | None = None
        self.evaluate_auth_type()

    def evaluate_auth_type(self) -> None:
        """
        Determine authentication type from available credentials.

        Evaluates credentials in priority order: EXCHANGE (personal access token),
        OAUTH2 (access + refresh tokens), ACCESS_TOKEN (access only), BASIC
        (username + password).
        """
        creds = self._credential_manager.get_credentials()
        self._auth_type = self._eval_auth_type(creds)

    @property
    def auth_type(self) -> str | None:
        """
        Get current authentication type.

        Returns
        -------
        str or None
            Current authentication type from AuthType enum.
        """
        return self._auth_type

    def is_refreshable(self) -> bool:
        """
        Check if current authentication supports token refresh.

        Returns True for OAUTH2 (refresh token) and EXCHANGE (personal access token),
        False for BASIC and ACCESS_TOKEN.

        Returns
        -------
        bool
            Whether authentication type supports refresh.
        """
        return self._auth_type in [AuthType.OAUTH2.value, AuthType.EXCHANGE.value]

    def get_auth_parameters(self, kwargs: dict) -> dict:
        """
        Add authentication headers/parameters to HTTP request kwargs.

        Adds Authorization Bearer header for token-based auth or auth tuple
        for basic authentication.

        Parameters
        ----------
        kwargs : dict
            HTTP request arguments to modify.

        Returns
        -------
        dict
            Modified kwargs with authentication parameters.
        """
        creds = self._credential_manager.get_credentials()
        if self._auth_type in (
            AuthType.EXCHANGE.value,
            AuthType.OAUTH2.value,
            AuthType.ACCESS_TOKEN.value,
        ):
            access_token = creds[CredentialsVars.DHCORE_ACCESS_TOKEN.value]
            kwargs = ensure_headers(**kwargs)
            kwargs["headers"]["Authorization"] = f"Bearer {access_token}"
        elif self._auth_type == AuthType.BASIC.value:
            user = creds[CredentialsVars.DHCORE_USER.value]
            password = creds[CredentialsVars.DHCORE_PASSWORD.value]
            kwargs["auth"] = (user, password)
        return kwargs

    def _eval_auth_type(self, creds: dict) -> str | None:
        """
        Determine authentication type from available credentials.

        Evaluates in priority order:

            EXCHANGE (personal access token)
            OAUTH2 (access + refresh tokens)
            ACCESS_TOKEN (access only)
            BASIC (username + password)
            None (no valid credentials type)

        Parameters
        ----------
        creds : dict
            Available credential values.

        Returns
        -------
        str or None
            Authentication type from AuthType enum, or None if no valid credentials.
        """
        if creds[CredentialsVars.DHCORE_PERSONAL_ACCESS_TOKEN.value] is not None:
            return AuthType.EXCHANGE.value
        if (
            creds[CredentialsVars.DHCORE_ACCESS_TOKEN.value] is not None
            and creds[CredentialsVars.DHCORE_REFRESH_TOKEN.value] is not None
        ):
            return AuthType.OAUTH2.value
        if creds[CredentialsVars.DHCORE_ACCESS_TOKEN.value] is not None:
            return AuthType.ACCESS_TOKEN.value
        if (
            creds[CredentialsVars.DHCORE_USER.value] is not None
            and creds[CredentialsVars.DHCORE_PASSWORD.value] is not None
        ):
            return AuthType.BASIC.value
        return None



# SPDX-FileCopyrightText: © 2025 DSLab - Fondazione Bruno Kessler
#
# SPDX-License-Identifier: Apache-2.0

from __future__ import annotations

import typing

from digitalhub.entities._commons.enums import EntityTypes
from digitalhub.entities._processors.processors import context_processor

if typing.TYPE_CHECKING:
    from digitalhub.entities.run._base.entity import Run


ENTITY_TYPE = EntityTypes.RUN.value


def new_run(
    project: str,
    kind: str,
    uuid: str | None = None,
    labels: list[str] | None = None,
    task: str | None = None,
    local_execution: bool = False,
    **kwargs,
) -> Run:
    """
    Create a new object.

    Parameters
    ----------
    project : str
        Project name.
    kind : str
        Kind the object.
    uuid : str
        ID of the object.
    labels : list[str]
        List of labels.
    task : str
        Name of the task associated with the run.
    local_execution : bool
        Flag to determine if object has local execution.
    **kwargs : dict
        Spec keyword arguments.

    Returns
    -------
    Run
        Object instance.

    Examples
    --------
    >>> obj = new_run(project="my-project",
    >>>               kind="python+run",
    >>>               task="task-string")
    """
    return context_processor.create_context_entity(
        project=project,
        kind=kind,
        uuid=uuid,
        labels=labels,
        task=task,
        local_execution=local_execution,
        entity_type=ENTITY_TYPE,
        **kwargs,
    )


def get_run(
    identifier: str,
    project: str | None = None,
) -> Run:
    """
    Get object from backend.

    Parameters
    ----------
    identifier : str
        Entity key (store://...) or entity ID.
    project : str
        Project name.

    Returns
    -------
    Run
        Object instance.

    Examples
    --------
    Using entity key:
    >>> obj = get_run("store://my-run-key")

    Using entity ID:
    >>> obj = get_run("my-run-id"
    >>>               project="my-project")
    """
    return context_processor.read_unversioned_entity(
        identifier=identifier,
        entity_type=ENTITY_TYPE,
        project=project,
    )


def list_runs(
    project: str,
    q: str | None = None,
    name: str | None = None,
    kind: str | None = None,
    user: str | None = None,
    state: str | None = None,
    created: str | None = None,
    updated: str | None = None,
    function: str | None = None,
    workflow: str | None = None,
    task: str | None = None,
    action: str | None = None,
) -> list[Run]:
    """
    List all latest version objects from backend.

    Parameters
    ----------
    project : str
        Project name.
    q : str
        Query string to filter objects.
    name : str
        Object name.
    kind : str
        Kind of the object.
    user : str
        User that created the object.
    state : str
        Object state.
    created : str
        Creation date filter.
    updated : str
        Update date filter.
    function : str
        Function key filter.
    workflow : str
        Workflow key filter.
    task : str
        Task string filter.
    action : str
        Action name filter.

    Returns
    -------
    list[Model]
        List of object instances.

    Examples
    --------
    >>> objs = list_runs(project="my-project")
    """
    return context_processor.list_context_entities(
        project=project,
        entity_type=ENTITY_TYPE,
        q=q,
        name=name,
        kind=kind,
        user=user,
        state=state,
        created=created,
        updated=updated,
        function=function,
        workflow=workflow,
        task=task,
        action=action,
    )


def import_run(
    file: str | None = None,
    key: str | None = None,
    reset_id: bool = False,
    context: str | None = None,
) -> Run:
    """
    Import an object from a YAML file or from a storage key.

    Parameters
    ----------
    file : str
        Path to the YAML file.
    key : str
        Entity key (store://...).
    reset_id : bool
        Flag to determine if the ID of executable entities should be reset.
    context : str
        Project name to use for context resolution.

    Returns
    -------
    Run
        Object instance.

    Example
    -------
    >>> obj = import_run("my-run.yaml")
    """
    return context_processor.import_context_entity(
        file,
        key,
        reset_id,
        context,
    )


def load_run(file: str) -> Run:
    """
    Load object from a YAML file and update an existing object into the backend.

    Parameters
    ----------
    file : str
        Path to YAML file.

    Returns
    -------
    Run
        Object instance.

    Examples
    --------
    >>> obj = load_run("my-run.yaml")
    """
    return context_processor.load_context_entity(file)


def update_run(entity: Run) -> Run:
    """
    Update object. Note that object spec are immutable.

    Parameters
    ----------
    entity : Run
        Object to update.

    Returns
    -------
    Run
        Entity updated.

    Examples
    --------
    >>> obj = update_run(obj)
    """
    return context_processor.update_context_entity(
        project=entity.project,
        entity_type=entity.ENTITY_TYPE,
        entity_id=entity.id,
        entity_dict=entity.to_dict(),
    )


def delete_run(
    identifier: str,
    project: str | None = None,
    entity_id: str | None = None,
) -> dict:
    """
    Delete object from backend.

    Parameters
    ----------
    identifier : str
        Entity key (store://...) or entity name.
    project : str
        Project name.
    entity_id : str
        Entity ID.

    Returns
    -------
    dict
        Response from backend.

    Examples
    --------
    >>> obj = delete_run("store://my-run-key")
    >>> obj = delete_run(
    ...     "my-run-id",
    ...     project="my-project",
    ... )
    """
    return context_processor.delete_context_entity(
        identifier=identifier,
        entity_type=ENTITY_TYPE,
        project=project,
        entity_id=entity_id,
        unversioned=True,
    )



# SPDX-FileCopyrightText: © 2025 DSLab - Fondazione Bruno Kessler
#
# SPDX-License-Identifier: Apache-2.0

from __future__ import annotations

import typing

from digitalhub.entities._base.versioned.entity import VersionedEntity
from digitalhub.entities._commons.enums import EntityTypes
from digitalhub.entities._processors.processors import context_processor

if typing.TYPE_CHECKING:
    from digitalhub.entities._base.entity.metadata import Metadata
    from digitalhub.entities.trigger._base.spec import TriggerSpec
    from digitalhub.entities.trigger._base.status import TriggerStatus


class Trigger(VersionedEntity):
    """
    A class representing a trigger.
    """

    ENTITY_TYPE = EntityTypes.TRIGGER.value

    def __init__(
        self,
        project: str,
        name: str,
        uuid: str,
        kind: str,
        metadata: Metadata,
        spec: TriggerSpec,
        status: TriggerStatus,
        user: str | None = None,
    ) -> None:
        super().__init__(project, name, uuid, kind, metadata, spec, status, user)
        self.spec: TriggerSpec
        self.status: TriggerStatus

    def stop(self) -> None:
        """
        Stop trigger.
        """
        return context_processor.stop_entity(self.project, self.ENTITY_TYPE, self.id)



# SPDX-FileCopyrightText: © 2025 DSLab - Fondazione Bruno Kessler
#
# SPDX-License-Identifier: Apache-2.0

from __future__ import annotations

import typing

from digitalhub.context.builder import context_builder
from digitalhub.entities._commons.enums import EntityTypes
from digitalhub.factory.entity import entity_factory
from digitalhub.stores.client.base.factory import get_client
from digitalhub.stores.client.common.enums import ApiCategories, BackendOperations
from digitalhub.utils.exceptions import ContextError, EntityNotExistsError

if typing.TYPE_CHECKING:
    from digitalhub.context.context import Context
    from digitalhub.entities.project._base.entity import Project


def build_context(project: Project, overwrite: bool = False) -> Context:
    """
    Build a new context for a project.

    Creates or updates a context instance for the given project
    in the global context registry.

    Parameters
    ----------
    project : Project
        The project object used to build the context.
    overwrite : bool
        If True, overwrites existing context if it exists. Default is False.

    Returns
    -------
    Context
        The newly created or existing context instance.
    """
    return context_builder.build(project, overwrite)


def get_context(project: str) -> Context:
    """
    Get the context for a given project name.

    Parameters
    ----------
    project : str
        Project name.

    Returns
    -------
    Context
        The context for the given project name.
    """
    try:
        return context_builder.get(project)
    except ContextError:
        try:
            return get_context_from_remote(project)
        except EntityNotExistsError as e:
            raise ContextError(f"Context '{project}' not found remotely nor locally.") from e


def get_context_from_remote(project: str) -> Context:
    """
    Fetch project context from remote backend and create local context.


    Parameters
    ----------
    project : str
        The name of the project to fetch from remote.

    Returns
    -------
    Context
        The context instance created from the remote project data.
    """
    try:
        client = get_client()
        api = client.build_api(
            ApiCategories.BASE.value,
            BackendOperations.READ.value,
            entity_type=EntityTypes.PROJECT.value,
            entity_name=project,
        )
        obj = client.read_object(api)
        entity_factory.build_entity_from_dict(obj)
        return context_builder.get(project)
    except EntityNotExistsError:
        raise ContextError(f"Project '{project}' not found.")


def delete_context(project: str) -> None:
    """
    Delete the context for a given project name.

    Parameters
    ----------
    project : str
        Project name.
    """
    context_builder.remove(project)



# SPDX-FileCopyrightText: © 2025 DSLab - Fondazione Bruno Kessler
#
# SPDX-License-Identifier: Apache-2.0

from __future__ import annotations

from digitalhub.entities._base.entity.spec import Spec, SpecValidator


class TriggerSpec(Spec):
    """
    TriggerSpec specifications.
    """

    def __init__(
        self,
        task: str,
        template: dict,
        function: str | None = None,
        workflow: str | None = None,
    ) -> None:
        super().__init__()
        self.task = task
        self.template = template
        self.function = function
        self.workflow = workflow


class TriggerValidator(SpecValidator):
    """
    TriggerValidator validator.
    """

    task: str
    """Task string."""

    template: dict
    """Template map."""

    function: str = None
    """Function string."""

    workflow: str = None
    """Workflow string."""



# SPDX-FileCopyrightText: © 2025 DSLab - Fondazione Bruno Kessler
#
# SPDX-License-Identifier: Apache-2.0

from __future__ import annotations

from pydantic import Field

from digitalhub.entities.trigger._base.spec import TriggerSpec, TriggerValidator

regexp = r"store://([^/]+)/(.+)"


class TriggerSpecLifecycle(TriggerSpec):
    """
    TriggerSpecLifecycle specifications.
    """

    def __init__(
        self,
        task: str,
        template: dict,
        function: str | None = None,
        workflow: str | None = None,
        key: str | None = None,
        states: list[str] | None = None,
    ) -> None:
        super().__init__(task, template, function, workflow)
        self.key = key
        self.states = states


class TriggerValidatorLifecycle(TriggerValidator):
    """
    TriggerValidatorLifecycle validator.
    """

    key: str | None = Field(pattern=regexp)
    """Entity key."""

    states: list[str] = None
    """List of entity states."""



# SPDX-FileCopyrightText: © 2025 DSLab - Fondazione Bruno Kessler
#
# SPDX-License-Identifier: Apache-2.0

from __future__ import annotations

import typing

from digitalhub.entities._base.versioned.entity import VersionedEntity
from digitalhub.entities._commons.enums import EntityTypes
from digitalhub.utils.generic_utils import decode_base64_string

if typing.TYPE_CHECKING:
    from digitalhub.entities._base.entity.metadata import Metadata
    from digitalhub.entities.log._base.spec import LogSpec
    from digitalhub.entities.log._base.status import LogStatus


class Log(VersionedEntity):
    """
    A class representing a log.
    """

    ENTITY_TYPE = EntityTypes.LOG.value

    def __init__(
        self,
        project: str,
        name: str,
        uuid: str,
        kind: str,
        metadata: Metadata,
        spec: LogSpec,
        status: LogStatus,
        user: str | None = None,
    ) -> None:
        super().__init__(project, name, uuid, kind, metadata, spec, status, user)
        self.spec: LogSpec
        self.status: LogStatus
        self._content: str | None = None
        self._text: str | None = None

    ##############################
    #  Log methods
    ##############################

    def set_content(self, content: str) -> None:
        """
        Set log content.

        Parameters
        ----------
        content : str
            Log content.
        """
        self._content = content
        self._text = decode_base64_string(content)

    @property
    def text(self) -> str | None:
        """
        Get log content as text.

        Returns
        -------
        str | None
            Log content as text.
        """
        return self._text



# SPDX-FileCopyrightText: © 2025 DSLab - Fondazione Bruno Kessler
#
# SPDX-License-Identifier: Apache-2.0

from __future__ import annotations

from digitalhub.entities.model._base.status import ModelStatus


class ModelStatusHuggingface(ModelStatus):
    """
    ModelStatusHuggingface status.
    """



# SPDX-FileCopyrightText: © 2025 DSLab - Fondazione Bruno Kessler
#
# SPDX-License-Identifier: Apache-2.0

from __future__ import annotations

import typing
from abc import abstractmethod

from digitalhub.entities._base._base.entity import Base

if typing.TYPE_CHECKING:
    from digitalhub.entities._base.entity.metadata import Metadata
    from digitalhub.entities._base.entity.spec import Spec
    from digitalhub.entities._base.entity.status import Status


class Entity(Base):
    """
    Abstract class for entities.

    An entity is a collection of metadata, specifications.and status
    representing a variety of objects handled by Digitalhub.
    """

    # Entity type
    # Need to be set in subclasses
    ENTITY_TYPE: str

    # Attributes to render as dict. Need to be expanded in subclasses.
    _obj_attr = ["kind", "metadata", "spec", "status", "user", "key"]

    def __init__(
        self,
        kind: str,
        metadata: Metadata,
        spec: Spec,
        status: Status,
        user: str | None = None,
    ) -> None:
        self.kind = kind
        self.metadata = metadata
        self.spec = spec
        self.status = status
        self.user = user

        # Need to be set in subclasses
        self.key: str

    @abstractmethod
    def save(self, update: bool = False) -> Entity:
        """
        Abstract save method.
        """

    @abstractmethod
    def refresh(self) -> Entity:
        """
        Abstract refresh method.
        """

    def _update_attributes(self, obj: Entity) -> None:
        """
        Update attributes.

        Parameters
        ----------
        obj : dict
            Mapping representation of object.
        """
        self.metadata = obj.metadata
        self.spec = obj.spec
        self.status = obj.status
        self.user = obj.user

    @abstractmethod
    def export(self) -> str:
        """
        Abstract export method.
        """

    def add_relationship(self, relation: str, dest: str, source: str | None = None) -> None:
        """
        Add relationship to entity metadata.

        Parameters
        ----------
        relation : str
            The type of relationship.
        dest : str
            The target entity.
        source : str
            The source entity.
        """
        if self.metadata.relationships is None:
            self.metadata.relationships = []
        obj = {"type": relation, "dest": dest}
        if source is not None:
            obj["source"] = source
        self.metadata.relationships.append(obj)

    def to_dict(self) -> dict:
        """
        Override default to_dict method to add the possibility to exclude
        some attributes. This requires to set a list of _obj_attr
        attributes in the subclass.

        Returns
        -------
        dict
            A dictionary containing the attributes of the entity instance.
        """
        return {k: v for k, v in super().to_dict().items() if k in self._obj_attr}

    def _post_create_hook_before_save(self) -> None:
        """
        Hook method called after the creation of the entity but before saving
        in Core.
        Can be overridden in subclasses to implement custom behavior.
        """

    def _post_read_hook(self) -> None:
        """
        Hook method called after reading the entity from Core.
        Can be overridden in subclasses to implement custom behavior.
        """

    def __repr__(self) -> str:
        """
        Return string representation of the entity object.

        Returns
        -------
        str
            A string representing the entity instance.
        """
        return str(self.to_dict())



# SPDX-FileCopyrightText: © 2025 DSLab - Fondazione Bruno Kessler
#
# SPDX-License-Identifier: Apache-2.0

from __future__ import annotations

from digitalhub.entities._base.entity.spec import Spec, SpecValidator
from digitalhub.entities.task._base.models import K8s


class RunSpec(Spec):
    """RunSpec specifications."""

    def __init__(
        self,
        task: str,
        local_execution: bool = False,
        function: str | None = None,
        workflow: str | None = None,
        volumes: list[dict] | None = None,
        resources: dict | None = None,
        envs: list[dict] | None = None,
        secrets: list[str] | None = None,
        profile: str | None = None,
        **kwargs,
    ) -> None:
        self.task = task
        self.local_execution = local_execution
        self.function = function
        self.workflow = workflow
        self.volumes = volumes
        self.resources = resources
        self.envs = envs
        self.secrets = secrets
        self.profile = profile


class RunValidator(SpecValidator, K8s):
    """
    RunValidator validator.
    """

    # Task parameters
    function: str | None = None
    workflow: str | None = None

    # Run parameters
    task: str
    """The task string associated with the run."""

    local_execution: bool = False
    """Flag to indicate if the run will be executed locally."""



# SPDX-FileCopyrightText: © 2025 DSLab - Fondazione Bruno Kessler
#
# SPDX-License-Identifier: Apache-2.0

from __future__ import annotations

from abc import abstractmethod
from typing import IO, Any


class DataframeReader:
    """
    Dataframe reader abstract class.
    """

    ##############################
    # Read methods
    ##############################

    @abstractmethod
    def read_df(self, path_or_buffer: str | IO, extension: str, **kwargs) -> Any:
        """
        Read DataFrame from path or buffer.
        """

    @abstractmethod
    def read_table(self, *args, **kwargs) -> Any:
        """
        Read table from db.
        """

    ##############################
    # Write methods
    ##############################

    @abstractmethod
    def write_df(self, df: Any, dst: Any, extension: str | None = None, **kwargs) -> None:
        """
        Method to write a dataframe to a file.
        """

    @staticmethod
    @abstractmethod
    def write_csv(df: Any, *args, **kwargs) -> str:
        """
        Write DataFrame as csv.
        """

    @staticmethod
    @abstractmethod
    def write_parquet(df: Any, *args, **kwargs) -> str:
        """
        Write DataFrame as parquet.
        """

    @staticmethod
    @abstractmethod
    def write_table(df: Any, *args, **kwargs) -> str:
        """
        Write DataFrame as table.
        """

    ##############################
    # Utils
    ##############################

    @staticmethod
    @abstractmethod
    def get_schema(df: Any) -> Any:
        """
        Get schema.
        """

    @staticmethod
    @abstractmethod
    def get_preview(df: Any) -> Any:
        """
        Get preview.
        """

    @staticmethod
    @abstractmethod
    def concat_dfs(dfs: list[Any]) -> Any:
        """
        Concatenate a list of DataFrames.
        """

    @staticmethod
    @abstractmethod
    def get_limit_arg_name() -> str:
        """
        Get limit argument name for read methods.
        """



# SPDX-FileCopyrightText: © 2025 DSLab - Fondazione Bruno Kessler
#
# SPDX-License-Identifier: Apache-2.0
from __future__ import annotations

import json
from functools import lru_cache
from pathlib import Path
from typing import Any

import pytest
from jsonschema import ValidationError, validate

from digitalhub.factory.entity import entity_factory

ROOT = Path("test/instances")
ENTITIES_PATH = ROOT / "entities"
SCHEMAS_PATH = ROOT / "schemas"


class SchemaRegistry:
    """Registry for JSON schemas with lazy loading."""

    def __init__(self, schemas_path: Path):
        self._schemas_path = schemas_path
        self._schema_paths = {path.stem: path for path in schemas_path.rglob("*.json")}

    @lru_cache(maxsize=None)
    def get_schema(self, kind: str) -> dict[str, Any]:
        """Load and cache schema by kind."""
        if kind not in self._schema_paths:
            raise ValueError(f"Schema not found for kind: {kind}")
        return json.loads(self._schema_paths[kind].read_text())

    @property
    def available_kinds(self) -> set[str]:
        """Return all available schema kinds."""
        return set(self._schema_paths.keys())


class EntityValidator:
    """Validator for entity JSON files."""

    def __init__(self, schema_registry: SchemaRegistry):
        self.schema_registry = schema_registry

    def load_entity(self, path: Path) -> dict[str, Any]:
        """Load entity from JSON file."""
        try:
            return json.loads(path.read_text())
        except json.JSONDecodeError as e:
            raise ValueError(f"Invalid JSON in {path}: {e}") from e

    def build_from_file(self, path: Path) -> tuple[dict[str, Any], str]:
        """Build entity object from JSON file using factory."""
        entity = self.load_entity(path)

        if "kind" not in entity:
            raise ValueError(f"Missing 'kind' field in {path}")
        if "spec" not in entity:
            raise ValueError(f"Missing 'spec' field in {path}")

        kind = entity["kind"]
        built = entity_factory.build_spec(kind, **entity["spec"])
        return built.to_dict(), kind

    def validate(self, built: dict[str, Any], kind: str) -> None:
        """Validate built object against its schema."""
        schema = self.schema_registry.get_schema(kind)
        try:
            validate(instance=built, schema=schema)
        except ValidationError as e:
            raise AssertionError(f"Validation failed for kind '{kind}': {e.message}") from e

    def check_schema_completeness(self, built: dict[str, Any], kind: str) -> list[str]:
        """Check if rebuilt object includes all schema properties."""
        schema = self.schema_registry.get_schema(kind)
        missing_fields = []

        if "properties" in schema:
            schema_fields = set(schema["properties"].keys())
            built_fields = set(built.keys())
            missing_fields = sorted(schema_fields - built_fields)

        return missing_fields


# Discover entity files with collision-resistant naming
def discover_entities(entities_path: Path) -> dict[str, Path]:
    """Discover all entity JSON files, handling name collisions gracefully."""
    entities = {}
    for path in entities_path.rglob("*.json"):
        # Use relative path from entities_path for uniqueness
        key = str(path.relative_to(entities_path))
        entities[key] = path
    return entities


# Initialize components
schema_registry = SchemaRegistry(SCHEMAS_PATH)
validator = EntityValidator(schema_registry)
entity_paths = discover_entities(ENTITIES_PATH)


class TestValidate:
    """Test entity JSON files build correctly and validate against schemas."""

    @pytest.mark.parametrize("entity_file", sorted(entity_paths.keys()))
    def test_entity_validation(self, entity_file: str):
        """Test that entity builds from JSON and validates against its schema."""
        path = entity_paths[entity_file]
        built, kind = validator.build_from_file(path)
        validator.validate(built, kind)

    @pytest.mark.parametrize("entity_file", sorted(entity_paths.keys()))
    def test_schema_field_completeness(self, entity_file: str):
        """Test that rebuilt object includes all fields defined in schema (required and optional)."""
        path = entity_paths[entity_file]
        built, kind = validator.build_from_file(path)

        missing = validator.check_schema_completeness(built, kind)
        if missing:
            schema = validator.schema_registry.get_schema(kind)
            required = schema.get("required", [])
            missing_required = [f for f in missing if f in required]
            missing_optional = [f for f in missing if f not in required]

            msg_parts = []
            if missing_required:
                msg_parts.append(f"Missing REQUIRED fields: {missing_required}")
            if missing_optional:
                msg_parts.append(f"Missing optional fields: {missing_optional}")

            pytest.fail(f"Schema completeness check failed for '{kind}': {'; '.join(msg_parts)}")



# SPDX-FileCopyrightText: © 2025 DSLab - Fondazione Bruno Kessler
#
# SPDX-License-Identifier: Apache-2.0

from __future__ import annotations

from digitalhub.entities._base._base.entity import Base


class Status(Base):
    """
    Base Status class.
    The status class contains information about the state of an entity,
    for example, the state of a RUNNING run, and eventual error message.
    """

    def __init__(
        self,
        state: str,
        message: str | None = None,
        transitions: list[dict] | None = None,
        k8s: dict | None = None,
        **kwargs,
    ) -> None:
        self.state = state
        self.message = message
        self.transitions = transitions
        self.k8s = k8s

        self._any_setter(**kwargs)

    @classmethod
    def from_dict(cls, obj: dict) -> Status:
        """
        Return entity status object from dictionary.

        Parameters
        ----------
        obj : dict
            A dictionary containing the attributes of the entity status.

        Returns
        -------
        EntityStatus
            An entity status object.
        """
        return cls(**obj)



# SPDX-FileCopyrightText: © 2025 DSLab - Fondazione Bruno Kessler
#
# SPDX-License-Identifier: Apache-2.0

from __future__ import annotations

from digitalhub.entities._base.entity.spec import Spec, SpecValidator
from digitalhub.entities.project._base.models import ProfileConfig


class ProjectSpec(Spec):
    """
    ProjectSpec specifications.
    """

    def __init__(
        self,
        source: str | None = None,
        functions: list | None = None,
        artifacts: list | None = None,
        workflows: list | None = None,
        dataitems: list | None = None,
        models: list | None = None,
        config: dict | None = None,
        **kwargs,
    ) -> None:
        self.source = source if source is not None else "./"
        self.functions = functions if functions is not None else []
        self.artifacts = artifacts if artifacts is not None else []
        self.workflows = workflows if workflows is not None else []
        self.dataitems = dataitems if dataitems is not None else []
        self.models = models if models is not None else []
        self.config = config if config is not None else {}


class ProjectValidator(SpecValidator):
    """
    ProjectValidator validator.
    """

    source: str | None = None
    """The project's source."""

    functions: list | None = None
    """List of project's functions."""

    artifacts: list | None = None
    """List of project's artifacts."""

    workflows: list | None = None
    """List of project's workflows."""

    dataitems: list | None = None
    """List of project's dataitems."""

    models: list | None = None
    """List of project's models."""

    config: ProfileConfig | None = None
    """Project's config."""



# SPDX-FileCopyrightText: © 2025 DSLab - Fondazione Bruno Kessler
#
# SPDX-License-Identifier: Apache-2.0

from __future__ import annotations

from digitalhub.entities.trigger._base.status import TriggerStatus


class TriggerStatusLifecycle(TriggerStatus):
    """
    TriggerStatusLifecycle status.
    """



# SPDX-FileCopyrightText: © 2025 DSLab - Fondazione Bruno Kessler
#
# SPDX-License-Identifier: Apache-2.0

from __future__ import annotations

import importlib
import pkgutil
import re
from types import ModuleType

from digitalhub.factory.enums import FactoryEnum


def import_module(package: str) -> ModuleType:
    """
    Import a module dynamically by package name.

    Parameters
    ----------
    package : str
        The fully qualified package name to import.

    Returns
    -------
    ModuleType
        The imported module object.

    Raises
    ------
    ModuleNotFoundError
        If the specified package cannot be found.
    RuntimeError
        If any other error occurs during import.
    """
    try:
        return importlib.import_module(package)
    except ModuleNotFoundError as e:
        raise ModuleNotFoundError(f"Package {package} not found.") from e
    except Exception as e:
        raise RuntimeError(f"An error occurred while importing {package}.") from e


def list_runtimes() -> list[str]:
    """
    List all installed DigitalHub runtime packages.

    Scans installed packages for those matching the pattern
    'digitalhub_runtime_*'.

    Returns
    -------
    list of str
        List of runtime package names.

    Raises
    ------
    RuntimeError
        If an error occurs while scanning for runtime packages.
    """
    try:
        runtimes: list[str] = []
        for _, name, _ in pkgutil.iter_modules():
            if re.match(FactoryEnum.RGX_RUNTIMES.value, name):
                runtimes.append(name)
        return runtimes
    except Exception as e:
        raise RuntimeError("Error listing installed runtimes.") from e



# SPDX-FileCopyrightText: © 2025 DSLab - Fondazione Bruno Kessler
#
# SPDX-License-Identifier: Apache-2.0

from __future__ import annotations

import typing
from warnings import warn

from digitalhub.entities._commons.enums import EntityKinds, EntityTypes
from digitalhub.entities._processors.processors import context_processor
from digitalhub.entities.model.huggingface.crud import log_huggingface
from digitalhub.entities.model.mlflow.crud import log_mlflow
from digitalhub.entities.model.model.crud import log_generic_model
from digitalhub.entities.model.sklearn.crud import log_sklearn

if typing.TYPE_CHECKING:
    from digitalhub.entities.model._base.entity import Model
    from digitalhub.utils.types import SourcesOrListOfSources

ENTITY_TYPE = EntityTypes.MODEL.value


def new_model(
    project: str,
    name: str,
    kind: str,
    uuid: str | None = None,
    description: str | None = None,
    labels: list[str] | None = None,
    embedded: bool = False,
    path: str | None = None,
    **kwargs,
) -> Model:
    """
    Create a new object.

    Parameters
    ----------
    project : str
        Project name.
    name : str
        Object name.
    kind : str
        Kind the object.
    uuid : str
        ID of the object.
    description : str
        Description of the object (human readable).
    labels : list[str]
        List of labels.
    embedded : bool
        Flag to determine if object spec must be embedded in project spec.
    path : str
        Object path on local file system or remote storage. It is also the destination path of upload() method.
    **kwargs : dict
        Spec keyword arguments.

    Returns
    -------
    Model
        Object instance.

    Examples
    --------
    >>> obj = new_model(project="my-project",
    >>>                    name="my-model",
    >>>                    kind="model",
    >>>                    path="s3://my-bucket/my-key")
    """
    return context_processor.create_context_entity(
        project=project,
        name=name,
        kind=kind,
        uuid=uuid,
        description=description,
        labels=labels,
        embedded=embedded,
        entity_type=ENTITY_TYPE,
        path=path,
        **kwargs,
    )


def log_model(
    project: str,
    name: str,
    kind: str,
    source: SourcesOrListOfSources,
    drop_existing: bool = False,
    path: str | None = None,
    **kwargs,
) -> Model:
    """
    Create and upload an object.

    Parameters
    ----------
    project : str
        Project name.
    name : str
        Object name.
    kind : str
        Kind the object.
    source : SourcesOrListOfSources
        Model location on local path.
    drop_existing : bool
        Whether to drop existing entity with the same name.
    path : str
        Destination path of the model. If not provided, it's generated.
    **kwargs : dict
        New model spec parameters.

    Returns
    -------
    Model
        Object instance.

    Examples
    --------
    >>> obj = log_model(project="my-project",
    >>>                 name="my-model",
    >>>                 kind="model",
    >>>                 source="./local-path")
    """
    warn(
        "log_model will in version 0.16 log a model of kind 'model'. "
        "To log a model of a specific kind, use the log_<kind> methods (e.g. log_sklearn, log_mlflow)."
    )
    kwargs["project"] = project
    kwargs["name"] = name
    kwargs["source"] = source
    kwargs["drop_existing"] = drop_existing
    kwargs["path"] = path
    if kind == EntityKinds.MODEL_HUGGINGFACE.value:
        return log_huggingface(**kwargs)
    elif kind == EntityKinds.MODEL_MLFLOW.value:
        return log_mlflow(**kwargs)
    elif kind == EntityKinds.MODEL_SKLEARN.value:
        return log_sklearn(**kwargs)
    return log_generic_model(**kwargs)


def get_model(
    identifier: str,
    project: str | None = None,
    entity_id: str | None = None,
) -> Model:
    """
    Get object from backend.

    Parameters
    ----------
    identifier : str
        Entity key (store://...) or entity name.
    project : str
        Project name.
    entity_id : str
        Entity ID.

    Returns
    -------
    Model
        Object instance.

    Examples
    --------
    Using entity key:
    >>> obj = get_model("store://my-model-key")

    Using entity name:
    >>> obj = get_model("my-model-name"
    >>>                 project="my-project",
    >>>                 entity_id="my-model-id")
    """
    return context_processor.read_context_entity(
        identifier=identifier,
        entity_type=ENTITY_TYPE,
        project=project,
        entity_id=entity_id,
    )


def get_model_versions(
    identifier: str,
    project: str | None = None,
) -> list[Model]:
    """
    Get object versions from backend.

    Parameters
    ----------
    identifier : str
        Entity key (store://...) or entity name.
    project : str
        Project name.

    Returns
    -------
    list[Model]
        List of object instances.

    Examples
    --------
    Using entity key:
    >>> objs = get_model_versions("store://my-model-key")

    Using entity name:
    >>> objs = get_model_versions("my-model-name",
    >>>                           project="my-project")
    """
    return context_processor.read_context_entity_versions(
        identifier=identifier,
        entity_type=ENTITY_TYPE,
        project=project,
    )


def list_models(
    project: str,
    q: str | None = None,
    name: str | None = None,
    kind: str | None = None,
    user: str | None = None,
    state: str | None = None,
    created: str | None = None,
    updated: str | None = None,
    versions: str | None = None,
) -> list[Model]:
    """
    List all latest version objects from backend.

    Parameters
    ----------
    project : str
        Project name.
    q : str
        Query string to filter objects.
    name : str
        Object name.
    kind : str
        Kind of the object.
    user : str
        User that created the object.
    state : str
        Object state.
    created : str
        Creation date filter.
    updated : str
        Update date filter.
    versions : str
        Object version, default is latest.

    Returns
    -------
    list[Model]
        List of object instances.

    Examples
    --------
    >>> objs = list_models(project="my-project")
    """
    return context_processor.list_context_entities(
        project=project,
        entity_type=ENTITY_TYPE,
        q=q,
        name=name,
        kind=kind,
        user=user,
        state=state,
        created=created,
        updated=updated,
        versions=versions,
    )


def import_model(
    file: str | None = None,
    key: str | None = None,
    reset_id: bool = False,
    context: str | None = None,
) -> Model:
    """
    Import an object from a YAML file or from a storage key.

    Parameters
    ----------
    file : str
        Path to the YAML file.
    key : str
        Entity key (store://...).
    reset_id : bool
        Flag to determine if the ID of executable entities should be reset.
    context : str
        Project name to use for context resolution.

    Returns
    -------
    Model
        Object instance.

    Examples
    --------
    >>> obj = import_model("my-model.yaml")
    """
    return context_processor.import_context_entity(
        file,
        key,
        reset_id,
        context,
    )


def load_model(file: str) -> Model:
    """
    Load object from a YAML file and update an existing object into the backend.

    Parameters
    ----------
    file : str
        Path to YAML file.

    Returns
    -------
    Model
        Object instance.

    Examples
    --------
    >>> obj = load_model("my-model.yaml")
    """
    return context_processor.load_context_entity(file)


def update_model(entity: Model) -> Model:
    """
    Update object. Note that object spec are immutable.

    Parameters
    ----------
    entity : Model
        Object to update.

    Returns
    -------
    Model
        Entity updated.

    Examples
    --------
    >>> obj = get_model("store://my-model-key")
    """
    return context_processor.update_context_entity(
        project=entity.project,
        entity_type=entity.ENTITY_TYPE,
        entity_id=entity.id,
        entity_dict=entity.to_dict(),
    )


def delete_model(
    identifier: str,
    project: str | None = None,
    entity_id: str | None = None,
    delete_all_versions: bool = False,
    cascade: bool = True,
) -> dict:
    """
    Delete object from backend.

    Parameters
    ----------
    identifier : str
        Entity key (store://...) or entity name.
    project : str
        Project name.
    entity_id : str
        Entity ID.
    delete_all_versions : bool
        Delete all versions of the named entity.
        If True, use entity name instead of entity key as identifier.
    cascade : bool
        Cascade delete.

    Returns
    -------
    dict
        Response from backend.

    Examples
    --------
    If delete_all_versions is False:
    >>> obj = delete_model("store://my-model-key")

    Otherwise:
    >>> obj = delete_model("my-model-name",
    >>>                    project="my-project",
    >>>                    delete_all_versions=True)
    """
    return context_processor.delete_context_entity(
        identifier=identifier,
        entity_type=ENTITY_TYPE,
        project=project,
        entity_id=entity_id,
        delete_all_versions=delete_all_versions,
        cascade=cascade,
    )



# SPDX-FileCopyrightText: © 2025 DSLab - Fondazione Bruno Kessler
#
# SPDX-License-Identifier: Apache-2.0

from __future__ import annotations

from typing import Any, Union

from pydantic import BaseModel, ValidationError

MetricType = Union[float, int, list[Union[float, int]]]


class Metric(BaseModel):
    """
    Pydantic model for validating metric values.

    This model ensures that metric values are of the correct type,
    accepting single numeric values or lists of numeric values.

    Attributes
    ----------
    value : MetricType
        The metric value, which can be a float, int, or list of floats/ints.
    """

    value: MetricType


def validate_metric_value(value: Any) -> MetricType:
    """
    Validate and convert a value to a proper metric type.

    Uses Pydantic validation to ensure the input value conforms to
    the MetricType specification (float, int, or list of floats/ints).

    Parameters
    ----------
    value : Any
        The value to validate and convert.

    Returns
    -------
    MetricType
        The validated metric value.

    Raises
    ------
    ValueError
        If the value cannot be converted to a valid metric type.
    """
    try:
        return Metric(value=value).value
    except ValidationError as e:
        raise ValueError("Invalid metric value. Must be a list of floats or ints or a float or an int.") from e


def set_metrics(
    metrics: dict[str, MetricType],
    key: str,
    value: Any,
    overwrite: bool,
    single_value: bool,
) -> dict[str, MetricType]:
    """
    Set or update a metric value in the metrics dictionary.

    This function routes to appropriate handling based on the value type
    and the single_value flag. It can handle single values, lists, and
    appending to existing metrics.

    Parameters
    ----------
    metrics : dict[str, MetricType]
        The metrics dictionary to update.
    key : str
        The metric key to set or update.
    value : Any
        The value to set for the metric.
    overwrite : bool
        Whether to overwrite existing metrics.
    single_value : bool
        Whether to treat the value as a single metric rather than
        appending to a list.

    Returns
    -------
    dict[str, MetricType]
        The updated metrics dictionary.
    """
    if isinstance(value, list):
        return handle_metric_list(metrics, key, value, overwrite)
    elif single_value:
        return handle_metric_single(metrics, key, value, overwrite)
    return handle_metric_list_append(metrics, key, value, overwrite)


def handle_metric_single(
    metrics: dict[str, MetricType],
    key: str,
    value: float | int,
    overwrite: bool,
) -> dict:
    """
    Handle setting a single metric value.

    Sets or overwrites a metric with a single numeric value. If the key
    already exists and overwrite is False, the existing value is preserved.

    Parameters
    ----------
    metrics : dict[str, MetricType]
        The metrics dictionary to update.
    key : str
        The metric key to set.
    value : float | int
        The single numeric value to set.
    overwrite : bool
        Whether to overwrite an existing metric with the same key.

    Returns
    -------
    dict
        The updated metrics dictionary.
    """
    if key not in metrics or overwrite:
        metrics[key] = value
    return metrics


def handle_metric_list_append(
    metrics: dict[str, MetricType],
    key: str,
    value: float | int,
    overwrite: bool,
) -> dict:
    """
    Handle appending a single value to a metric list.

    If the metric doesn't exist or overwrite is True, creates a new list
    with the single value. If the metric exists as a list, appends to it.
    If the metric exists as a single value, converts it to a list and appends.

    Parameters
    ----------
    metrics : dict[str, MetricType]
        The metrics dictionary to update.
    key : str
        The metric key to append to.
    value : float | int
        The numeric value to append.
    overwrite : bool
        Whether to overwrite an existing metric instead of appending.

    Returns
    -------
    dict
        The updated metrics dictionary.
    """
    if key not in metrics or overwrite:
        metrics[key] = [value]
    elif isinstance(metrics[key], list):
        metrics[key].append(value)
    else:
        metrics[key] = [metrics[key], value]
    return metrics


def handle_metric_list(
    metrics: dict[str, MetricType],
    key: str,
    value: list[int | float],
    overwrite: bool,
) -> dict:
    """
    Handle setting or extending a metric with a list of values.

    If the metric doesn't exist or overwrite is True, sets the metric to
    the provided list. If the metric exists and overwrite is False, extends
    the existing list with the new values.

    Parameters
    ----------
    metrics : dict[str, MetricType]
        The metrics dictionary to update.
    key : str
        The metric key to set or extend.
    value : list[int | float]
        The list of numeric values to set or extend with.
    overwrite : bool
        Whether to overwrite an existing metric instead of extending it.

    Returns
    -------
    dict
        The updated metrics dictionary.
    """
    if key not in metrics or overwrite:
        metrics[key] = value
    else:
        metrics[key].extend(value)
    return metrics



# SPDX-FileCopyrightText: © 2025 DSLab - Fondazione Bruno Kessler
#
# SPDX-License-Identifier: Apache-2.0

from __future__ import annotations

from enum import Enum


class SetCreds(Enum):
    """
    Supported credential environments.
    """

    DEFAULT = "__default"
    DH_PROFILE = "DH_NAME"


class ConfigurationVars(Enum):
    """
    List of supported configuration variables.
    """

    # S3
    S3_ENDPOINT_URL = "AWS_ENDPOINT_URL"
    S3_REGION = "AWS_REGION"
    S3_SIGNATURE_VERSION = "S3_SIGNATURE_VERSION"
    S3_PATH_STYLE = "S3_PATH_STYLE"

    # SQL
    DB_HOST = "DB_HOST"
    DB_PORT = "DB_PORT"
    DB_DATABASE = "DB_DATABASE"
    DB_PLATFORM = "DB_PLATFORM"
    DB_PG_SCHEMA = "DB_SCHEMA"

    # DHCORE
    DHCORE_ENDPOINT = "DHCORE_ENDPOINT"
    DHCORE_ISSUER = "DHCORE_ISSUER"
    DHCORE_WORKFLOW_IMAGE = "DHCORE_WORKFLOW_IMAGE"
    DHCORE_CLIENT_ID = "DHCORE_CLIENT_ID"
    DEFAULT_FILES_STORE = "DHCORE_DEFAULT_FILES_STORE"

    # OAUTH2
    OAUTH2_TOKEN_ENDPOINT = "OAUTH2_TOKEN_ENDPOINT"


class CredentialsVars(Enum):
    """
    List of supported credential variables.
    """

    # S3
    S3_ACCESS_KEY_ID = "AWS_ACCESS_KEY_ID"
    S3_SECRET_ACCESS_KEY = "AWS_SECRET_ACCESS_KEY"
    S3_SESSION_TOKEN = "AWS_SESSION_TOKEN"
    S3_CREDENTIALS_EXPIRATION = "AWS_CREDENTIALS_EXPIRATION"

    # SQL
    DB_USERNAME = "DB_USERNAME"
    DB_PASSWORD = "DB_PASSWORD"

    # DHCORE
    DHCORE_USER = "DHCORE_USER"
    DHCORE_PASSWORD = "DHCORE_PASSWORD"
    DHCORE_ACCESS_TOKEN = "DHCORE_ACCESS_TOKEN"
    DHCORE_REFRESH_TOKEN = "DHCORE_REFRESH_TOKEN"
    DHCORE_PERSONAL_ACCESS_TOKEN = "DHCORE_PERSONAL_ACCESS_TOKEN"



# SPDX-FileCopyrightText: © 2025 DSLab - Fondazione Bruno Kessler
#
# SPDX-License-Identifier: Apache-2.0

from __future__ import annotations

import time
import typing

from digitalhub.entities._base.metrics.entity import MetricsEntity
from digitalhub.entities._base.unversioned.entity import UnversionedEntity
from digitalhub.entities._commons.enums import EntityTypes, State
from digitalhub.entities._processors.processors import context_processor
from digitalhub.factory.entity import entity_factory
from digitalhub.factory.runtime import runtime_factory
from digitalhub.utils.exceptions import EntityError
from digitalhub.utils.logger import LOGGER

if typing.TYPE_CHECKING:
    from digitalhub.entities._base.entity.metadata import Metadata
    from digitalhub.entities.log._base.entity import Log
    from digitalhub.entities.run._base.spec import RunSpec
    from digitalhub.entities.run._base.status import RunStatus
    from digitalhub.runtimes._base import Runtime


class Run(UnversionedEntity, MetricsEntity):
    """
    A class representing a run.
    """

    ENTITY_TYPE = EntityTypes.RUN.value

    def __init__(
        self,
        project: str,
        uuid: str,
        kind: str,
        metadata: Metadata,
        spec: RunSpec,
        status: RunStatus,
        user: str | None = None,
    ) -> None:
        super().__init__(project, uuid, kind, metadata, spec, status, user)

        self.spec: RunSpec
        self.status: RunStatus

    ##############################
    #  Run Methods
    ##############################

    def build(self) -> None:
        """
        Build run.
        """
        executable = self._get_executable()
        task = self._get_task()
        new_spec = self._get_runtime().build(executable, task, self.to_dict())
        self.spec = entity_factory.build_spec(self.kind, **new_spec)
        self.status.state = State.BUILT.value
        self.save(update=True)

    def run(self) -> Run:
        """
        Run run.

        Returns
        -------
        Run
            Run object.
        """
        self.refresh()

        self.start_execution()
        self._setup_execution()

        try:
            status = self._get_runtime().run(self.to_dict())

        # Handle exceptions and set run status and message
        except Exception as e:
            self.refresh()
            if self.spec.local_execution:
                self.status.state = State.ERROR.value
            self.status.message = str(e)
            self.save(update=True)
            raise e

        # Unset run in context
        finally:
            self.end_execution()

        self.refresh()
        if not self.spec.local_execution:
            status.pop("state", None)
        new_status = {**self.status.to_dict(), **status}
        self.set_status(new_status)
        self.save(update=True)
        return self

    def wait(self, log_info: bool = True) -> Run:
        """
        Wait for run to finish.

        Parameters
        ----------
        log_info : bool
            If True, log information.

        Returns
        -------
        Run
            Run object.
        """
        start = time.time()
        while True:
            if log_info:
                LOGGER.info(f"Waiting for run {self.id} to finish...")
            self.refresh()
            time.sleep(5)
            if self.status.state in [
                State.STOPPED.value,
                State.ERROR.value,
                State.COMPLETED.value,
            ]:
                if log_info:
                    current = time.time() - start
                    LOGGER.info(f"Run {self.id} finished in {current:.2f} seconds.")
                return self

    def logs(self) -> list[Log]:
        """
        Get run logs. If no logs are present, an empty list is returned.
        In case of local execution, logs are printed to console.

        Returns
        -------
        list[Log]
            List of run logs.
        """
        return context_processor.read_run_logs(self.project, self.ENTITY_TYPE, self.id)

    def stop(self) -> None:
        """
        Stop run.
        """
        return context_processor.stop_entity(self.project, self.ENTITY_TYPE, self.id)

    def resume(self) -> None:
        """
        Resume run.
        """
        return context_processor.resume_entity(self.project, self.ENTITY_TYPE, self.id)

    ##############################
    #  Helpers
    ##############################

    def _setup_execution(self) -> None:
        """
        Setup run execution.
        """

    def start_execution(self) -> None:
        """
        Start run execution.
        """
        self._context().set_run(self)
        if self.spec.local_execution:
            # Check run state
            if self.status.state not in (State.BUILT.value, State.STOPPED.value):
                raise EntityError("Run is not in a state to run.")

            self.status.state = State.RUNNING.value
            self.save(update=True)

    def end_execution(self) -> None:
        """
        End run execution.
        """
        self._context().unset_run()

    def set_status(self, status: dict) -> None:
        """
        Set run status.

        Parameters
        ----------
        status : dict
            Status to set.
        """
        self.status: RunStatus = entity_factory.build_status(self.kind, **status)

    def _get_runtime(self) -> Runtime:
        """
        Build runtime to build run or execute it.

        Returns
        -------
        Runtime
            Runtime object.
        """
        return runtime_factory.build_runtime(self.kind, self.project)

    def _get_executable(self) -> dict:
        """
        Get executable object from backend. Reimplemented to avoid
        circular imports.

        Returns
        -------
        dict
            Executable (function or workflow) from backend.
        """
        exec_kind = entity_factory.get_executable_kind(self.kind)
        exec_type = entity_factory.get_entity_type_from_kind(exec_kind)
        string_to_split = getattr(self.spec, exec_type)
        exec_name, exec_id = string_to_split.split("://")[-1].split("/")[-1].split(":")
        return context_processor.read_context_entity(
            exec_name,
            entity_type=exec_type,
            project=self.project,
            entity_id=exec_id,
        ).to_dict()

    def _get_task(self) -> dict:
        """
        Get object from backend. Reimplemented to avoid
        circular imports.

        Returns
        -------
        dict
            Task from backend.
        """
        task_id = self.spec.task.split("://")[-1].split("/")[-1]
        return context_processor.read_unversioned_entity(
            task_id,
            entity_type=EntityTypes.TASK.value,
            project=self.project,
        ).to_dict()
